package api

import (
	"strconv"
	"strings"

	"git.autistici.org/ale/djrandom/util/tokenizer"
)

// MultiValuedMap is a map of attributes to multiple values.
type MultiValuedMap map[string][]string

// Add an attribute/value pair to the document.
func (m MultiValuedMap) Add(key, value string) {
	a, ok := m[key]
	if !ok {
		a = make([]string, 0)
	}
	m[key] = append(a, value)
}

// Document is an entry in the index.
type Document MultiValuedMap

func (d Document) Add(key, value string) {
	MultiValuedMap(d).Add(key, value)
}

func toTokens(s string) []string {
	return tokenizer.TokenFilter(
		tokenizer.Tokenize(strings.ToLower(s)),
		tokenizer.IsKeyword)
}

// ToDoc extracts the indexable tokens from a Song object.
func (song *Song) ToDoc() Document {
	doc := make(Document)

	maybeAdd := func(key string, value string, literal bool) {
		if value != "" {
			for _, t := range toTokens(value) {
				doc.Add(key, t)
			}
			if literal {
				doc.Add(key, "="+value)
			}
		}
	}

	maybeAdd("title", song.Meta.Title, true)
	maybeAdd("artist", song.Meta.Artist, true)
	maybeAdd("album", song.Meta.Album, true)
	maybeAdd("genre", song.Meta.Genre, false)
	maybeAdd("year", strconv.Itoa(song.Meta.Year), false)

	// Add (unique) mime types for each audio file.
	mimeTypesMap := make(map[string]struct{})
	for _, af := range song.Files {
		mimeTypesMap[af.MimeType] = struct{}{}
	}
	for mimeType, _ := range mimeTypesMap {
		doc.Add("mime_type", mimeType)
	}

	return doc
}
