package api

import (
	"math/rand"
)

type Transition struct {
	Src string
	Dst string
}

// FreqDist contains a probability distribution for a specific N-gram
// in the Markov model.  We store counters instead of probabilities so
// that we can update the map incrementally.
type FreqDist struct {
	D   map[string]int
	Sum int64
}

func NewFreqDist() *FreqDist {
	return &FreqDist{
		D:   make(map[string]int),
		Sum: 0,
	}
}

func (fd *FreqDist) Incr(song string) {
	counter, ok := fd.D[song]
	if ok {
		counter++
	} else {
		fd.D[song] = 1
	}
	fd.Sum++
}

func (fd *FreqDist) GetSong() string {
	t := rand.Int63n(fd.Sum)
	var cur int64 = 0
	for song, count := range fd.D {
		cur += int64(count)
		if cur > t {
			return song
		}
	}

	// This should be unreachable.
	return ""
}

// MarkovMap contains probability distributions for playlists.
// The input tuple size is 1, and it's hard-coded in the interface.
type MarkovMap struct {
	M map[string]*FreqDist
}

func NewMarkovMap() *MarkovMap {
	return &MarkovMap{
		M: make(map[string]*FreqDist),
	}
}

func (m *MarkovMap) GetSong(cur string) (string, bool) {
	dist, ok := m.M[cur]
	if !ok {
		return "", false
	}
	return dist.GetSong(), true
}

func (m *MarkovMap) Update(in chan Transition) {
	for t := range in {
		dist, ok := m.M[t.Src]
		if !ok {
			dist = NewFreqDist()
			m.M[t.Src] = dist
		}
		dist.Incr(t.Dst)
	}
}
