package main

import (
	"crypto/x509"
	"flag"
	"fmt"
	"io"
	"io/ioutil"
	"log"
	"net/url"
	"strings"

	"git.autistici.org/ale/djrandom/api"
	"git.autistici.org/ale/djrandom/util"
	"os/exec"
	"os"
)

var (
	serverUrl     = flag.String("server", "https://djrandom.incal.net", "Server URL")
	authKeyId     = flag.String("auth-key", "", "API authentication key")
	authKeySecret = flag.String("auth-secret", "", "API authentication secret")
	caFile        = flag.String("ca", "ca.pem", "CA X509 certificate (PEM-encoded)")
	player        = flag.String("player", "avplay -vn -nodisp -", "Audio player program")
)

func loadCA() *x509.CertPool {
	pool := x509.NewCertPool()
	data, err := ioutil.ReadFile(*caFile)
	if err != nil {
		return nil
	}
	pool.AppendCertsFromPEM(data)
	return pool
}

func addAuthToUrl(u string) string {
	pfxcount := 8
	if strings.HasPrefix(u, "http:") {
		pfxcount = 7
	}
	return fmt.Sprintf("%s%s:%s@%s", u[:pfxcount], *authKeyId, *authKeySecret, u[pfxcount:])
}

func play(client *util.HttpClient, song *api.Song) {
	if len(song.Files) == 0 {
		return
	}

	resp, err := client.GetRaw("/dl/" + song.Files[0].MD5)
	if err != nil {
		log.Fatal(err)
	}
	defer resp.Body.Close()

	if resp.StatusCode != 200 {
		log.Printf("error in request: status = %d", resp.StatusCode)
		return
	}

	playCmd := strings.Split(*player, " ")
	cmd := exec.Command(playCmd[0], playCmd[1:]...)
	cmd.Stdout = os.Stderr
	cmd.Stderr = os.Stderr
	stdin, err := cmd.StdinPipe()
	if err != nil {
		log.Fatal(err)
	}
	if err := cmd.Start(); err != nil {
		log.Fatal(err)
	}

	n, err := io.Copy(stdin, resp.Body)
	if err != nil {
		log.Printf("error copying audio: %v", err)
	}
	log.Printf("%d bytes copied", n)
	stdin.Close()

	// Ignore errors.
	cmd.Wait()
}

func search(query string) {
	authKey := &api.AuthKey{
		KeyId:  *authKeyId,
		Secret: *authKeySecret,
	}
	client := util.NewHttpClient(*serverUrl, authKey, loadCA())

	values := url.Values{}
	values.Add("q", query)
	req, _ := client.NewRequest("POST", "/api/search", strings.NewReader(values.Encode()))
	req.Header.Set("Content-Type", "application/x-www-form-urlencoded")

	var resp api.SearchResponse
	if err := client.DoJSON(nil, req, &resp); err != nil {
		log.Fatal(err)
	}

	for _, song := range resp.Results {
		fmt.Printf("%s %s\n", song.Meta.Artist, song.Meta.Title)
		play(client, song)
	}
}

func main() {
	flag.Parse()

	query := strings.Join(flag.Args(), " ")
	search(query)
}
