package main

import (
	"flag"
	"log"
	"net/http"
	"strings"
	"time"

	_ "net/http/pprof"

	"git.autistici.org/ale/djrandom/api"
	"git.autistici.org/ale/djrandom/client/upload"
	"git.autistici.org/ale/djrandom/util"
)

var (
	config     = util.NewConfig()
	configFile = flag.String("config", util.ExpandTilde("~/.djrandom.conf"), "Config file location")
	once       = flag.Bool("once", false, "Run once and then exit")
	httpAddr   = flag.String("debug-http", ":4040", "Port to run the debug HTTP server on")

	// Default parameters for the file scanner.
	defaultIncludePatterns = "*.flac,*.mp3"
	defaultMimeTypes       = []string{
		api.MIMETYPE_FLAC,
		api.MIMETYPE_MP3,
	}
)

func runScan(up *upload.Uploader, db *upload.FileDb) {
	musicDir := util.ExpandTilde(config.MustGetString("music_dir"))
	log.Printf("starting scan on %s", musicDir)

	// Create the filesystem scanner and configure it to select
	// files with a '.flac' extension.
	fs := upload.NewFSScanner(musicDir)
	fs.AddExcludePattern(".*")
	for _, p := range strings.Split(config.GetString("include-patterns", defaultIncludePatterns), ",") {
		fs.AddIncludePattern(p)
	}

	// Checking the db is cheaper than running 'file', so we put
	// the database filter first. The MimeType filter verifies
	// that the files are actually media files (so we don't blindly
	// trust the file extension).
	filter := upload.NewSimpleFilter(
		upload.IsNonEmptyFile,
		upload.NewFileDbFilter(
			db,
			upload.NewSimpleFilter(
				upload.NewMimeTypeFilterFunc(defaultMimeTypes),
				upload.NewSimpleFilter(upload.IsFileUnchanged, nil))))

	fs.WalkWithFilters(filter, up.SinkFunc())
}

func main() {
	flag.Parse()

	if err := config.Read(*configFile); err != nil {
		log.Printf("Warning: could not read config file %s: %s", *configFile, err.Error())
	}

	authKey := api.AuthKey{
		KeyId:  config.MustGetString("auth_key"),
		Secret: config.MustGetString("auth_secret"),
	}

	db := upload.NewFileDb(config.GetString("db_path", util.ExpandTilde("~/.djrandom.db")))
	httpClient := util.NewHttpClient(config.MustGetString("server_url"), &authKey, nil)
	up := upload.NewUploader(db, httpClient, 0, 4, upload.GetFingerprint)
	go up.Run()

	http.Handle("/", up)
	go func() {
		log.Fatal(http.ListenAndServe(*httpAddr, nil))
	}()

	for {
		runScan(up, db)
		if *once {
			// Sleep 60 seconds just to flush some stuff
			time.Sleep(time.Second * 60)
			break
		}
		time.Sleep(time.Hour * 3)
	}
}
