package upload

import (
	"log"
	"os"
	"path/filepath"
)

// getAbsPath finds a directory's absolute path by cd'ing into it and
// calling Getwd.
func getAbsPath(path string) (string, error) {
	oldCwd, oldErr := os.Getwd()
	err := os.Chdir(path)
	if err != nil {
		return path, err
	}
	abs, err := os.Getwd()
	if err != nil {
		return abs, err
	}
	if oldErr == nil {
		os.Chdir(oldCwd)
	}
	return abs, nil
}

// FSScanner represents a filesystem scan that can include or exclude
// specific patterns.
type FSScanner struct {
	Root            string
	excludePatterns []string
	includePatterns []string
	hasExcludes     bool
	hasIncludes     bool
}

func NewFSScanner(root string) *FSScanner {
	absPath, err := getAbsPath(root)
	if err != nil {
		absPath = root
	}

	return &FSScanner{
		Root:            absPath,
		excludePatterns: make([]string, 0),
		includePatterns: make([]string, 0),
		hasExcludes:     false,
		hasIncludes:     false,
	}
}

func (fs *FSScanner) AddExcludePattern(pat string) {
	fs.excludePatterns = append(fs.excludePatterns, pat)
	fs.hasExcludes = true
}

func (fs *FSScanner) AddIncludePattern(pat string) {
	fs.includePatterns = append(fs.includePatterns, pat)
	fs.hasIncludes = true
}

func (fs *FSScanner) match(path string, info os.FileInfo, err error, out chan *FileInfo) error {
	if err != nil {
		return nil
	}
	if fs.hasExcludes && MatchAny(info.Name(), fs.excludePatterns) {
		if info.IsDir() {
			return filepath.SkipDir
		} else {
			return nil
		}
	}
	if fs.hasIncludes && !MatchAny(info.Name(), fs.includePatterns) {
		return nil
	}
	out <- NewFileInfo(path, info)
	return nil
}

func (fs *FSScanner) Walk(out chan *FileInfo) error {
	return filepath.Walk(fs.Root, func(path string, info os.FileInfo, err error) error {
		return fs.match(path, info, err, out)
	})
}

// SinkFunc is a function that processes its input.
type SinkFunc func(*FileInfo)

// Filter is the generic interface for filters.
type Filter interface {
	Matches(*FileInfo) bool
}

// SimpleFilter lets you wrap a plain function with a Filter interface.
type SimpleFilter struct {
	next Filter
	fn   func(*FileInfo) bool
}

func NewSimpleFilter(fn func(*FileInfo) bool, next Filter) *SimpleFilter {
	return &SimpleFilter{next: next, fn: fn}
}

func (f *SimpleFilter) Matches(fileinfo *FileInfo) bool {
	ok := f.fn(fileinfo)
	if !ok {
		return false
	}
	if f.next != nil {
		return f.next.Matches(fileinfo)
	}
	return ok
}

// A simple filter that selects files with size greater than 0.
func IsNonEmptyFile(fileinfo *FileInfo) bool {
	return !fileinfo.Info.IsDir() && fileinfo.Info.Size() > 0
}

// WalkWithFilters scans the filesystem and applies the specified
// filters to the files that are found.
func (fs *FSScanner) WalkWithFilters(filter Filter, sink SinkFunc) {
	ch := make(chan *FileInfo)
	go func() {
		fs.Walk(ch)
		close(ch)
	}()

	log.Printf("starting scan of %s", fs.Root)
	count := 0
	matches := 0
	for fileinfo := range ch {
		count++
		if filter.Matches(fileinfo) {
			matches++
			sink(fileinfo)
		}
	}
	log.Printf("scan terminated, %d entries found, %d matching", count, matches)
}
