package upload

import (
	"io"
	"log"
	"net/http"
	"os"
	"strings"
	"sync"
	"time"

	"git.autistici.org/ale/djrandom/api"
	"git.autistici.org/ale/djrandom/util"
)

type Uploader struct {
	db            *FileDb
	limiter       *RateLimiter
	http          *util.HttpClient
	inCh          chan *FileInfo
	uploadCh      chan *FileInfo
	checkBufSize  int
	nProcs        int
	fingerprintFn func(*FileInfo) (string, error)

	// Some counters.
	NumProcessed  int
	NumUploaded   int
	NumDuplicates int
	NumPending    int
}

// Keep track of what the upload workers are doing, for debugging purposes.
var (
	uploaderState     = make(map[int]string)
	uploaderStateLock sync.Mutex
)

func setUploadWorkerState(num int, state string) {
	uploaderStateLock.Lock()
	defer uploaderStateLock.Unlock()
	uploaderState[num] = state
}

func getUploadWorkerState() map[int]string {
	uploaderStateLock.Lock()
	defer uploaderStateLock.Unlock()
	stateCopy := make(map[int]string)
	for idx, state := range uploaderState {
		stateCopy[idx] = state
	}
	return stateCopy
}

func NewUploader(db *FileDb, httpClient *util.HttpClient, rateLimit int, nProcs int, fingerprintFn func(*FileInfo) (string, error)) *Uploader {
	u := new(Uploader)
	u.db = db
	u.http = httpClient
	u.inCh = make(chan *FileInfo)
	u.uploadCh = make(chan *FileInfo)
	u.checkBufSize = 100
	u.nProcs = nProcs
	u.fingerprintFn = fingerprintFn
	if rateLimit > 0 {
		u.limiter = NewRateLimiter(rateLimit * 1024)
	}
	return u
}

func (u *Uploader) ServeHTTP(w http.ResponseWriter, r *http.Request) {
	io.WriteString(w, "debug\n")
}

func (u *Uploader) SinkFunc() func(*FileInfo) {
	return func(fileinfo *FileInfo) {
		log.Printf("queuing for upload: %s", fileinfo.Path)
		u.NumPending++
		u.inCh <- fileinfo
	}
}

// Run the uploader pipeline.
func (u *Uploader) Run() {
	for i := 0; i < u.nProcs; i++ {
		go u.uploadWorker(i)
	}
	u.checkWorker()
}

func (u *Uploader) checkWorker() {
	buf := make([]*FileInfo, u.checkBufSize)
	pos := 0

	// Read files from 'in', group them in chunks and poll the
	// server API to verify if they need to be uploaded.  Querying
	// the API in chunks saves one round-trip time per file, which
	// is critical if we're rebuilding the local database. Since
	// the input channel never closes, we don't want to leave
	// entries in the buffer pending for an undeterminate time, so
	// we set a timeout for flushing.
	rt := util.NewRestartableTimer()
	timeout := time.Duration(time.Second * 10)

	for {
		select {

		case fileinfo := <-u.inCh:
			buf[pos] = fileinfo
			pos++
			if pos == u.checkBufSize {
				u.checkFiles(buf[:pos])
				pos = 0
			}
			rt.Reset(timeout)

		case <-rt.C:
			log.Println("timeout expired, polling...")
			if pos > 0 {
				u.checkFiles(buf[:pos])
				pos = 0
			}
		}
	}
}

func (u *Uploader) checkFingerprintsApi(fingerprints []string) ([]string, []string, error) {
	log.Printf("Checking fingerprints: %s", strings.Join(fingerprints, ","))
	req := api.FingerprintRequest{Fingerprints: fingerprints}
	var resp api.FingerprintResponse
	err := u.http.Post("/api/check_fp", &req, &resp)
	if err != nil {
		log.Printf("Error checking fingerprints: %s", err.Error())
		return []string{}, []string{}, err
	}
	return resp.Missing, resp.Dupes, nil
}

// Check all files in buffer with an API call.  All missing files will
// be sent to the 'uploadCh' channel. All files will be marked as seen
// in the db.
func (u *Uploader) checkFiles(infos []*FileInfo) {
	// Compute fingerprints of all files. Also, build a map of
	// path -> fileinfo, since we're going to have to use it later
	// on.
	fingerprints := []string{}
	infoMap := make(map[string]*FileInfo)
	for _, info := range infos {
		fp, err := u.fingerprintFn(info)
		if err != nil {
			log.Printf("Error computing fingerprint of %s: %s", info.Path, err.Error())
			// Fingerprint errors are permanent.
			u.db.Update(info)
			continue
		}
		info.Meta["fingerprint"] = fp
		infoMap[fp] = info
		fingerprints = append(fingerprints, fp)
	}

	// Do we actually have any fingerprints to check? Cut down on useless
	// empty queries.
	if len(fingerprints) == 0 {
		return
	}

	// Perform the RPC query, and get two lists of fingerprints as
	// a result.
	missingFp, dupesFp, err := u.checkFingerprintsApi(fingerprints)
	if err != nil {
		log.Printf("Remote error checking fingerprints: %s", err.Error())
		return
	}

	// Missing files are sent to the uploader stage.
	for _, fp := range missingFp {
		u.uploadCh <- infoMap[fp]
	}

	// Duplicate files are marked as seen in the db.
	u.NumDuplicates += len(dupesFp)
	for _, fp := range dupesFp {
		u.db.Update(infoMap[fp])
	}

	return
}

// UploadWorker handles uploading files that come in via the
// 'uploadCh' channel.
func (u *Uploader) uploadWorker(num int) {
	setUploadWorkerState(num, "idle")
	for fileinfo := range u.uploadCh {
		u.NumPending--
		u.NumProcessed++
		setUploadWorkerState(num, "uploading "+fileinfo.Path)
		log.Printf("Uploading %s", fileinfo.Path)
		err := u.upload(fileinfo)
		if err != nil {
			log.Printf("Error uploading %s: %s", fileinfo.Path, err.Error())
		} else {
			u.db.Update(fileinfo)
			u.NumUploaded++
		}
		setUploadWorkerState(num, "idle")
	}
}

func (u *Uploader) upload(fileinfo *FileInfo) error {
	var resp api.UploadResponse
	file, err := os.Open(fileinfo.Path)
	if err != nil {
		return err
	}
	defer file.Close()
	stat, _ := file.Stat()
	return u.http.Put("/upload", file, stat.Size(), &resp)
}
