package partition

import (
	"errors"
	"fmt"
	"html/template"
	"net/http"
	"net/rpc"
	"strings"
	"sync"

	"git.autistici.org/ale/djrandom/util/status"
)

const debugText = `<html>
<body>
<title>Partition Map: {{.Name}}</title>
<table>
 <thead>
  <tr>
   <th>EndKey</th><th>Target</th>
  </tr>
 </thead>
 <tbody>
  {{range .Pmap.P}}
  <tr>
   <td><tt>{{.EndKey | printf "%v" }}</tt></td>
   <td><a href="{{.Target | httpize}}">{{.Target}}</a></td>
  </tr>
  {{end}}
 </tbody>
</table>
<p>Version: <tt>{{.Pmap.GetVersion}}</tt></p>
</body>
</html>`

const debugListText = `<html>
<body>
<title>Partition Maps</title>
<p>
  {{range .}}
  <a href="/debug/part/{{.}}">{{.}}</a><br>
  {{end}}
</p>
</body>
</html>`

func httpize(s string) string {
	if !strings.HasPrefix(s, "http") {
		return fmt.Sprintf("http://%s", s)
	}
	return s
}

var debug = template.Must(template.New("PartitionMap debug").Funcs(template.FuncMap{"httpize": httpize}).Parse(debugText))
var debugList = template.Must(template.New("PartitionMap debug list").Parse(debugListText))

var debugPartitionMap = make(map[string]*PartitionMap)
var debugPartitionMapLock sync.Mutex

func registerPartitionMap(name string, pmap *PartitionMap) {
	debugPartitionMapLock.Lock()
	defer debugPartitionMapLock.Unlock()
	debugPartitionMap[name] = pmap
}

type debugRPC struct{}

type GetVersionRequest struct {
	Partition string
}

type GetVersionResponse struct {
	Partition string
	Version   string
}

func (server debugRPC) GetVersion(req *GetVersionRequest, resp *GetVersionResponse) error {
	debugPartitionMapLock.Lock()
	pmap, ok := debugPartitionMap[req.Partition]
	debugPartitionMapLock.Unlock()
	if !ok {
		return errors.New("Not found")
	}
	resp.Partition = req.Partition
	resp.Version = pmap.GetVersion()
	return nil
}

type debugHTTP struct{}

func (server debugHTTP) ServeHTTP(w http.ResponseWriter, r *http.Request) {
	name := r.URL.Path

	var err error
	if name == "" {
		pmapList := make([]string, 0, len(debugPartitionMap))
		debugPartitionMapLock.Lock()
		for name, _ := range debugPartitionMap {
			pmapList = append(pmapList, name)
		}
		debugPartitionMapLock.Unlock()
		err = debugList.Execute(w, pmapList)
	} else {
		debugPartitionMapLock.Lock()
		pmap, ok := debugPartitionMap[name]
		debugPartitionMapLock.Unlock()
		if !ok {
			http.Error(w, "Not Found", http.StatusNotFound)
			return
		}
		ctx := struct {
			Pmap *PartitionMap
			Name string
		}{pmap, name}
		err = debug.Execute(w, &ctx)
	}

	if err != nil {
		fmt.Fprintln(w, "rpc: error executing template: ", err.Error())
	}
}

var debugRpcServer = &debugRPC{}
var debugHttpServer = &debugHTTP{}

func init() {
	http.Handle("/debug/part/",
		http.StripPrefix("/debug/part/", debugHttpServer))
	rpc.RegisterName("Partition", debugRpcServer)
	status.RegisterDebugEndpoint("/debug/part/")
}

func (pmap *PartitionMap) HandleDebug(name string) {
	registerPartitionMap(name, pmap)
}
