// Expose a WebDAV interface to the song database.

package frontend

import (
	"errors"
	"fmt"
	"log"
	"net/http"
	"path/filepath"
	"strings"

	"git.autistici.org/ale/djrandom/api"
	"git.autistici.org/ale/djrandom/services"
	"git.autistici.org/ale/djrandom/util"
	"git.autistici.org/ale/djrandom/util/webdav"
)

var (
	davNotFound = errors.New("Not Found")
)

const (
	davRootKey = "@@ROOT@@"
)

// DavFs implements a filesystem in terms of the webdavi.Fs
// interface. It exports a "/artist/album/song_title"
// filesystem hierarchy.
type DavFs struct {
	db      services.Database
	index   services.Index
	storage services.Storage
}

func NewDavFs(db services.Database, index services.Index, storage services.Storage) *DavFs {
	return &DavFs{db, index, storage}
}

func parseDavPath(path string) ([]string, error) {
	if path == "" {
		return []string{}, nil
	}
	parts := strings.Split(path, "/")
	if len(parts) > 3 {
		return nil, davNotFound
	}
	return parts, nil
}

// Find a song given artist / album / title.
func (davfs *DavFs) findSong(pathParts []string) *api.Song {
	// Remove the extension from the song title (which appears
	// as the file name).
	ext := filepath.Ext(pathParts[2])
	title := pathParts[2][:len(pathParts[2])-len(ext)]

	// This is a mess from the point of view of encodings, it
	// will only work some of the time...
	query := fmt.Sprintf("author:\"=%s\" artist:\"=%s\" title:\"=%s\"", pathParts[0], pathParts[1], title)

	// Search for an exact match. If more than one, just pick the first.
	songIdMap := davfs.index.Search(query)
	if len(songIdMap) == 0 {
		return nil
	}
	log.Printf("findSong: %v -> %v", query, songIdMap)
	var songId string
	for id, _ := range songIdMap {
		songId = id
		break
	}

	// Retrieve the song from the database and pick an AudioFile.
	song, ok := davfs.db.GetSong(nil, songId)
	if !ok {
		return nil
	}
	return song
}

func (davfs *DavFs) Get(w http.ResponseWriter, r *http.Request) {
	path := strings.Trim(r.URL.Path, "/")
	pathParts, err := parseDavPath(path)
	if err != nil || len(pathParts) != 3 {
		http.Error(w, "Not Found", http.StatusNotFound)
		return
	}

	// Build an exact match query.
	song := davfs.findSong(pathParts)
	if song == nil {
		http.Error(w, "Not Found", http.StatusNotFound)
		return
	}
	audioFileKey := song.Files[0].MD5

	// Directly call the storage reverse proxy handler. We need
	// to wipe clean the Authorization header in the original
	// request though, or the storage client authentication won't
	// work.
	r.Header.Del("Authorization")
	downloadProxyHandler(davfs.storage, audioFileKey, w, r)
}

func (davfs *DavFs) Stat(path string) (webdav.Attrs, error) {
	pathParts, err := parseDavPath(path)
	if err != nil {
		return webdav.Attrs{}, err
	}

	log.Printf("DavFs.Stat: parts=%v (%d)", pathParts, len(pathParts))

	// A Stat request does not return a name as it always refers to
	// the request URL.
	attrs := webdav.Attrs{}
	if len(pathParts) == 3 {
		// find song by artist / album / title
		song := davfs.findSong(pathParts)
		if song == nil {
			return attrs, errors.New("Not found")
		}
		attrs.Size = song.Files[0].Size
		attrs.Mtime = song.Info.CreatedAt
		attrs.Ctime = song.Info.CreatedAt
	} else {
		attrs.IsDir = true
	}

	return attrs, nil
}

func (davfs *DavFs) Ls(path string) ([]webdav.Attrs, error) {
	pathParts, err := parseDavPath(path)
	if err != nil {
		return nil, err
	}
	log.Printf("DavFs.Ls: parts=%v (%d)", pathParts, len(pathParts))

	var entries []webdav.Attrs
	switch len(pathParts) {
	case 0:
		// Root. Return list of authors.
		entries = davfs.readAuthors()
	case 1:
		// Author, list all albums.
		entries = davfs.readAlbumsForAuthor(pathParts[0])
	case 2:
		// Author + album, list songs.
		entries = davfs.readSongsForAlbum(pathParts[0], pathParts[1])
	default:
		return nil, errors.New("Not a directory")
	}

	// Return.
	return entries, nil
}

func makeDirList(entries []string) []webdav.Attrs {
	attrs := make([]webdav.Attrs, 0, len(entries))
	for _, e := range entries {
		attrs = append(attrs, webdav.Attrs{
			Name:  e,
			IsDir: true,
		})
	}
	return attrs
}

func (davfs *DavFs) readAuthors() []webdav.Attrs {
	var allAuthors []string
	if err := davfs.db.Get(nil, "davfs", davRootKey, &allAuthors); err != nil {
		return []webdav.Attrs{}
	}
	return makeDirList(allAuthors)
}

func (davfs *DavFs) readAlbumsForAuthor(author string) []webdav.Attrs {
	var albums []string
	if err := davfs.db.Get(nil, "davfs", author, &albums); err != nil {
		log.Printf("Error retrieving albums for '%s': %s", author, err)
		return []webdav.Attrs{}
	}
	log.Printf("albums for author '%s': %v", author, albums)
	return makeDirList(albums)
}

func getExtensionForMimeType(mimeType string) string {
	switch mimeType {
	case api.MIMETYPE_MP3:
		return ".mp3"
	case api.MIMETYPE_FLAC:
		return ".flac"
	}
	return ""
}

func (davfs *DavFs) readSongsForAlbum(author, album string) []webdav.Attrs {
	// Build an exact match query.
	query := fmt.Sprintf("author:\"=%s\" album:\"=%s\"", author, album)

	songIdMap := davfs.index.Search(query)
	songIds := make([]string, 0, len(songIdMap))
	for songId, _ := range songIdMap {
		songIds = append(songIds, songId)
	}

	// Fetch objects from the database in parallel.
	s, err := davfs.db.NewSession()
	if err != nil {
		return nil
	}
	defer s.Close()
	rCh := util.RunAtOnce(numParallelQueries, songIds, func(songId string) interface{} {
		if song, ok := davfs.db.GetSong(s, songId); ok {
			return song
		}
		return nil
	})

	// Read all results into the rpc response.
	attrs := make([]webdav.Attrs, 0, len(songIds))
	for i := 0; i < len(songIds); i++ {
		song := (<-rCh).(*api.Song)
		if song == nil {
			continue
		}
		// file := song.GetBestAudioFile()
		// if file == nil {
		//	continue
		// }
		file := song.Files[0]
		fileName := fmt.Sprintf("%s%s", song.Meta.Title, getExtensionForMimeType(file.MimeType))
		attrs = append(attrs, webdav.Attrs{
			Name:  fileName,
			Size:  file.Size,
			Mtime: song.Info.CreatedAt,
			Ctime: song.Info.CreatedAt,
		})
	}
	return attrs
}
