// Front-end HTTP server for djrandom.

package main

import (
	"flag"
	"log"
	"net/rpc"

	"git.autistici.org/ale/djrandom/server/frontend"
	"git.autistici.org/ale/djrandom/server/frontend/albumart"
	"git.autistici.org/ale/djrandom/services/database"
	"git.autistici.org/ale/djrandom/services/index"
	"git.autistici.org/ale/djrandom/services/storage"
	"git.autistici.org/ale/djrandom/services/tasks"
	"git.autistici.org/ale/djrandom/util"
	"git.autistici.org/ale/djrandom/util/daemon"
)

var (
	config     = util.NewConfig()
	configFile = flag.String("config", "/etc/djrandom/server.conf", "Config file location")
	addr       = flag.String("addr", ":3001", "TCP address to listen on.")
)

func main() {
	flag.Parse()

	if err := config.Read(*configFile); err != nil {
		log.Fatalf("Error: could not read config file %s: %s", *configFile, err.Error())
	}

	daemon.EnableStatusPage = false
	daemon.Setup()

	// Create client stubs for all our services.
	dbImpl := database.NewDatabaseImplFromConfig(config)
	storageClient := storage.NewDistributedStorageClientFromConfig(config)
	indexClient := index.NewIndexClientFromConfig(config)
	taskClient := tasks.NewTaskClientFromConfig(config, dbImpl)
	aaCache := albumart.NewAlbumArtCache(dbImpl, config.MustGetString("last_fm_api_key"))

	// Bundle everything into an AppContext.
	ac := frontend.AppContext{
		Db:         dbImpl,
		Storage:    storageClient,
		TaskClient: taskClient,
		Index:      indexClient,
		AlbumArt:   aaCache,
		XsrfSecret: config.GetString("xsrf-secret", "test"),
	}

	frontend.RegisterService(&ac)

	// Start the RPC HTTP handler for debug functions.
	rpc.HandleHTTP()

	log.Printf("Starting HTTP front-end server on %s", *addr)
	err := util.ListenAndServe(*addr, nil)
	if err != nil {
		log.Fatal("ListenAndServe: ", err)
	}
}
