package frontend

import (
	"errors"
	"log"
	"net/http"
	"time"

	"git.autistici.org/ale/djrandom/api"
)

func Login(w http.ResponseWriter, r *djRequest) {
	var ctx struct {
		Email        string
		ErrorMessage string
		XsrfToken    string
	}

	// Attempt to authenticate the user.
	if r.Request.Method == "POST" && r.ValidateXsrf() {
		email := r.Request.FormValue("email")
		password := r.Request.FormValue("password")
		user, err := r.Ctx.Authenticate(email, password)
		if err != nil {
			ctx.Email = email
			ctx.ErrorMessage = "Authentication Failed"
		} else {
			doLogin(w, r, user)
			// Redirect to post-auth page...
			w.Header().Set("Location", "/")
			w.WriteHeader(http.StatusFound)
			return
		}
	}
	ctx.XsrfToken = r.NewXsrfToken(w)

	r.Ctx.RenderTemplate(w, "login.html", &ctx)
}

type alertMsg struct {
	Msg  string
	Type string
}

func UserPage(w http.ResponseWriter, r *djRequest) {
	msgs := make([]alertMsg, 0)

	user, ok := r.Ctx.Db.GetUser(nil, r.AuthUser)
	if !ok || user == nil {
		log.Printf("UserPage(%s): error retrieving user", r.AuthUser)
		http.Error(w, "Internal Error", http.StatusInternalServerError)
		return
	}

	// Perform any user-requested actions.
	if r.Request.Method == "POST" && r.ValidateXsrf() {
		var okMsg string
		var err error
		formId := r.Request.FormValue("form_id")
		switch {
		case formId == "change_pw":
			err = changePassword(user, r)
			okMsg = "Password changed successfully"
		case formId == "create_api_key":
			err = createApiKey(user, r)
			okMsg = "Created new API key"
		case formId == "invite":
			err = invite(user, r)
			okMsg = "Invite sent"
		default:
			http.Error(w, err.Error(), http.StatusBadRequest)
			return
		}
		if err != nil {
			msgs = append(msgs, alertMsg{err.Error(), "error"})
		} else {
			msgs = append(msgs, alertMsg{okMsg, "info"})
		}
	}

	// Build the template context.
	ctx := struct {
		*templateContext
		User      *api.User
		AuthKeys  []*api.AuthKey
		Alerts    []alertMsg
		XsrfToken string
	}{newTemplateContext("user", r), user, getUserAuthKeys(user, r), msgs, r.NewXsrfToken(w)}

	r.Ctx.RenderTemplate(w, "user.html", &ctx)
}

func getUserAuthKeys(user *api.User, r *djRequest) []*api.AuthKey {
	s, err := r.Ctx.Db.NewSession()
	if err != nil {
		return nil
	}
	out := make([]*api.AuthKey, 0, len(user.AuthKeyIds))
	for _, keyId := range user.AuthKeyIds {
		if key, ok := r.Ctx.Db.GetAuthKey(s, keyId); ok {
			out = append(out, key)
		}
	}
	return out
}

func invite(user *api.User, r *djRequest) error {
	email := r.Request.FormValue("email")
	if email == "" {
		return errors.New("No email provided")
	}
	if user.InvitesLeft <= 0 {
		return errors.New("No more invites left")
	}

	// Create a new user in 'pending' state.
	newUser := &api.User{
		Email:       email,
		CreatedAt:   time.Now().Unix(),
		InvitedBy:   user.Email,
		InvitesLeft: 3,
		State:       "pending",
	}
	err := r.Ctx.Db.PutUser(nil, newUser)
	if err != nil {
		return err
	}

	// TODO: send activation email!

	// Update the number of invites left for the current user.
	// Note that this is a test-and-set operation without a lock
	// on the object that is being modified, so it is inherently
	// unreliable.
	//
	// Also, if there is a database error here, log it but do not
	// cause the request to fail (since the invite has already
	// been sent).
	user.InvitesLeft--
	r.Ctx.Db.PutUser(nil, user)
	return nil
}

func changePassword(user *api.User, r *djRequest) error {
	password := r.Request.FormValue("password")
	passwordConfirm := r.Request.FormValue("password_confirm")
	if password == "" || passwordConfirm == "" {
		return errors.New("Passwords must not be empty")
	}
	if password != passwordConfirm {
		return errors.New("You typed two different passwords")
	}
	if len(password) < 6 {
		return errors.New("Passwords must be at least 6 characters")
	}

	user.SetPassword(password)
	return r.Ctx.Db.PutUser(nil, user)
}

func createApiKey(user *api.User, r *djRequest) error {
	if err := r.Ctx.Db.PutAuthKey(nil, user.NewAuthKey()); err != nil {
		return err
	}
	return r.Ctx.Db.PutUser(nil, user)
}
