// MR job that builds the metadata index for the DAV filesystem view.

package main

import (
	"flag"
	"log"

	"github.com/dgryski/dmrgo"

	"git.autistici.org/ale/djrandom/services"
	"git.autistici.org/ale/djrandom/services/database"
	"git.autistici.org/ale/djrandom/util"
)

var (
	config     = util.NewConfig()
	configFile = flag.String("config", "/etc/djrandom/server.conf", "Config file location")
)

const (
	davRootKey = "@@ROOT@@"
)

type MRDavFs struct {
	db    services.Database
	proto dmrgo.StreamProtocol
}

// Map emits data for each hierarchical level of the resulting filesystem.
func (mr *MRDavFs) Map(key string, value string, emitter dmrgo.Emitter) {
	songId := value
	song, ok := mr.db.GetSong(nil, songId)
	if !ok {
		log.Printf("song %s does not exist", songId)
		return
	}

	emit := func(k, v string) {
		kv := mr.proto.MarshalKV(k, v)
		emitter.Emit(kv.Key, kv.Value)
	}

	if song.Meta.Artist != "" {
		if song.Meta.Album != "" {
			emit(song.Meta.Artist, song.Meta.Album)
		}
		emit(davRootKey, song.Meta.Artist)
	}
}

func (mr *MRDavFs) MapFinal(emitter dmrgo.Emitter) {
}

func uniquify(in []string) []string {
	h := make(map[string]struct{})
	for _, s := range in {
		h[s] = struct{}{}
	}
	i := 0
	for s, _ := range h {
		in[i] = s
		i++
	}
	return in[:i]
}

// Reduce collects statistics into a counter
func (mr *MRDavFs) Reduce(key string, values []string, emitter dmrgo.Emitter) {
	var svalues []string
	mr.proto.UnmarshalKVs(key, values, &key, &svalues)

	svalues = uniquify(svalues)
	log.Printf("%s: %d values", key, len(svalues))
	mr.db.Put(nil, "davfs", key, svalues)
}

func main() {
	flag.Parse()

	if err := config.Read(*configFile); err != nil {
		log.Fatalf("Error: could not read config file %s: %s", *configFile, err.Error())
	}

	dbImpl := database.NewDatabaseImplFromConfig(config)
	mr := &MRDavFs{
		db:    dbImpl,
		proto: &dmrgo.JSONProtocol{},
	}
	dmrgo.Main(mr)
}
