package main

import (
	"flag"
	"log"
	"regexp"
	"strings"

	"github.com/dgryski/dmrgo"

	"git.autistici.org/ale/djrandom/services"
	"git.autistici.org/ale/djrandom/services/database"
	"git.autistici.org/ale/djrandom/util"
)

var (
	config     = util.NewConfig()
	configFile = flag.String("config", "/etc/djrandom/server.conf", "Config file location")

	// Pattern for metadata normalization.
	nonWordPattern = regexp.MustCompile("\\W+")
)

type MRDedupMeta struct {
	db    services.Database
	sess  services.Session
	proto dmrgo.StreamProtocol
}

// Deduplication is performed on "normalized" artist/title pairs.
type keyType struct {
	Artist string
	Title  string
}

func normalize(s string) string {
	return strings.ToUpper(nonWordPattern.ReplaceAllString(s, ""))
}

// Map emits (artist/title, song_id) tuples.
func (mr *MRDedupMeta) Map(key string, value string, emitter dmrgo.Emitter) {
	songId := value
	song, ok := mr.db.GetSong(mr.sess, songId)
	if !ok {
		log.Printf("song %s does not exist", songId)
		return
	}

	k := keyType{
		Artist: normalize(song.Meta.Artist),
		Title:  normalize(song.Meta.Title),
	}
	kv := mr.proto.MarshalKV(&k, songId)
	emitter.Emit(kv.Key, kv.Value)
}

func (mr *MRDedupMeta) MapFinal(emitter dmrgo.Emitter) {
}

// Reduce de-duplicates identical songs.
func (mr *MRDedupMeta) Reduce(key string, values []string, emitter dmrgo.Emitter) {
	var songIds []string
	var k keyType
	mr.proto.UnmarshalKVs(key, values, &k, &songIds)

	log.Printf("Reduce(%v): %v", k, songIds)

	// Deduplicate.
	if len(songIds) < 2 {
		return
	}

	// kv := mr.proto.MarshalKV(user, m)
	// emitter.Emit(kv.Key, kv.Value)
}

func main() {
	flag.Parse()

	if err := config.Read(*configFile); err != nil {
		log.Fatalf("Error: could not read config file %s: %s", *configFile, err.Error())
	}

	dbImpl := database.NewDatabaseImplFromConfig(config)
	dbSess, err := dbImpl.NewSession()
	if err != nil {
		log.Fatalf("Error: connecting to the db: %s", err)
	}
	defer dbSess.Close()

	mr := &MRDedupMeta{
		db:    dbImpl,
		sess:  dbSess,
		proto: &dmrgo.JSONProtocol{},
	}
	dmrgo.Main(mr)
}
