package main

import (
	"flag"
	"log"
	"sort"

	"github.com/dgryski/dmrgo"

	"git.autistici.org/ale/djrandom/api"
	"git.autistici.org/ale/djrandom/services"
	"git.autistici.org/ale/djrandom/services/database"
	"git.autistici.org/ale/djrandom/util"
)

var (
	config     = util.NewConfig()
	configFile = flag.String("config", "/etc/djrandom/server.conf", "Config file location")
)

const (
	globalUser = "@GLOBAL"
)

type MRMarkov struct {
	db    services.Database
	sess  services.Session
	proto dmrgo.StreamProtocol
}

// Map emits (user, playlog entry) tuples.
func (mr *MRMarkov) Map(key string, value string, emitter dmrgo.Emitter) {
	plKey := value
	log.Printf("PlayLogEntry(%s)", plKey)
	pl, ok := mr.db.GetPlayLog(mr.sess, plKey)
	if !ok {
		log.Printf("playlog entry %s does not exist", plKey)
		return
	}

	// Emit the transition for the global map.
	kv := mr.proto.MarshalKV(globalUser, pl)
	emitter.Emit(kv.Key, kv.Value)

	// Emit the transition for the user's map (unless the playlog
	// entry has a nil user, in case it comes from an old data import).
	if pl.User != "" {
		kv := mr.proto.MarshalKV(pl.User, pl)
		emitter.Emit(kv.Key, kv.Value)
	}
}

func (mr *MRMarkov) MapFinal(emitter dmrgo.Emitter) {
}

type PlayLogList []*api.PlayLogEntry

func (pl PlayLogList) Len() int      { return len(pl) }
func (pl PlayLogList) Swap(i, j int) { pl[i], pl[j] = pl[j], pl[i] }
func (pl PlayLogList) Less(i, j int) bool {
	return pl[i].Timestamp < pl[j].Timestamp
}

// Reduce generates per-user Markov maps.
func (mr *MRMarkov) Reduce(key string, values []string, emitter dmrgo.Emitter) {
	var plvalues []*api.PlayLogEntry
	var user string
	mr.proto.UnmarshalKVs(key, values, &user, &plvalues)
	log.Printf("Reduce(%s): %d entries", user, len(plvalues))

	sort.Sort(PlayLogList(plvalues))

	tc := make(chan api.Transition)
	go func() {
		for _, entry := range plvalues {
			var src, dst string
			n := len(entry.Songs)
			if n > 1 {
				src = entry.Songs[n-2]
				dst = entry.Songs[n-1]
			} else if n == 1 {
				src = ""
				dst = entry.Songs[0]
			} else {
				continue
			}
			tc <- api.Transition{Src: src, Dst: dst}
		}
		close(tc)
	}()

	m := api.NewMarkovMap()
	m.Update(tc)
	log.Printf("Markov map for %s = %v", user, m)

	// Store the new map in the db.
	if err := mr.db.PutMarkovMap(mr.sess, user, m); err != nil {
		log.Printf("Could not store map: %s", err)
	}

	// Emit it anyway for now.
	kv := mr.proto.MarshalKV(user, m)
	emitter.Emit(kv.Key, kv.Value)
}

func main() {
	flag.Parse()

	if err := config.Read(*configFile); err != nil {
		log.Fatalf("Error: could not read config file %s: %s", *configFile, err.Error())
	}

	dbImpl := database.NewDatabaseImplFromConfig(config)
	dbSess, err := dbImpl.NewSession()
	if err != nil {
		log.Fatalf("Error: connecting to the db: %s", err)
	}
	defer dbSess.Close()

	mr := &MRMarkov{
		db:    dbImpl,
		sess:  dbSess,
		proto: &dmrgo.JSONProtocol{},
	}
	dmrgo.Main(mr)
}
