// Compute metadata statistics for the song database.
//
package main

import (
	"flag"
	"fmt"
	"log"

	"github.com/dgryski/dmrgo"

	"git.autistici.org/ale/djrandom/services"
	"git.autistici.org/ale/djrandom/services/database"
	"git.autistici.org/ale/djrandom/util"
)

var (
	config     = util.NewConfig()
	configFile = flag.String("config", "/etc/djrandom/server.conf", "Config file location")
)

type MRStats struct {
	db    services.Database
	proto dmrgo.StreamProtocol
}

type keyType struct {
	Stat string
	Key  string
}

// Map emits statistics for a number of song attributes.
func (mr *MRStats) Map(key string, value string, emitter dmrgo.Emitter) {
	songId := value
	song, ok := mr.db.GetSong(nil, songId)
	if !ok {
		log.Printf("song %s does not exist", songId)
		return
	}

	emitStat := func(k, v string) {
		kv := mr.proto.MarshalKV(&keyType{Stat: k, Key: v}, 1)
		emitter.Emit(kv.Key, kv.Value)
	}

	// Song metadata.
	if song.Meta.Artist != "" {
		emitStat("artist", song.Meta.Artist)
	}
	if song.Meta.Genre != "" {
		emitStat("genre", song.Meta.Genre)
	}
	if song.Meta.Year > 0 {
		emitStat("year", fmt.Sprintf("%d", song.Meta.Year))
	}
	// Audio file metadata.
	for _, af := range song.Files {
		emitStat("mime_type", af.MimeType)
		if af.BitRate > 0 {
			emitStat("bitrate", fmt.Sprintf("%d", af.BitRate))
		}
	}
}

func (mr *MRStats) MapFinal(emitter dmrgo.Emitter) {
}

// Reduce collects statistics into a counter
func (mr *MRStats) Reduce(key string, values []string, emitter dmrgo.Emitter) {
	var k keyType
	var ivalues []int
	mr.proto.UnmarshalKVs(key, values, &k, &ivalues)

	c := 0
	for _, count := range ivalues {
		c += count
	}

	log.Printf("counter/%s/%s = %d", k.Stat, k.Key, c)
	kv := mr.proto.MarshalKV(&k, c)
	emitter.Emit(kv.Key, kv.Value)
}

func main() {
	flag.Parse()

	if err := config.Read(*configFile); err != nil {
		log.Fatalf("Error: could not read config file %s: %s", *configFile, err.Error())
	}

	dbImpl := database.NewDatabaseImplFromConfig(config)
	mr := &MRStats{
		db:    dbImpl,
		proto: &dmrgo.JSONProtocol{},
	}
	dmrgo.Main(mr)
}
