// Verifies that the index contents match what's stored in the
// database. It's a full sync, so it will remove elements from the
// index which are not in the db anymore (or whose status isn't
// "active").
//
// Right now this scans the entire index at once, which won't scale
// very well if the number of songs grows too large.

package main

import (
	"flag"
	"log"

	"github.com/dgryski/dmrgo"

	"git.autistici.org/ale/djrandom/services"
	"git.autistici.org/ale/djrandom/services/database"
	"git.autistici.org/ale/djrandom/services/index"
	"git.autistici.org/ale/djrandom/util"
)

var (
	config     = util.NewConfig()
	configFile = flag.String("config", "/etc/djrandom/server.conf", "Config file location")
	rebuild    = flag.Bool("rebuild", false, "Fully rebuild the index")
)

type MRVerify struct {
	db    services.Database
	idx   services.Index
	proto dmrgo.StreamProtocol
}

func sliceToMap(s []string) map[string]struct{} {
	m := make(map[string]struct{})
	for _, value := range s {
		m[value] = struct{}{}
	}
	return m
}

func (mr *MRVerify) addSong(songId string) {
	song, ok := mr.db.GetSong(nil, songId)
	if !ok {
		log.Printf("Error: song %s does not exist?!?", songId)
		return
	}

	if err := mr.idx.AddDoc(songId, song.ToDoc()); err != nil {
		log.Printf("Error: indexing %s: %s", songId, err)
	}
}

func (mr *MRVerify) deleteSong(songId string) {
	mr.idx.DeleteDoc(songId)
}

// Map emits statistics for a number of song attributes.
func (mr *MRVerify) Map(key string, value string, emitter dmrgo.Emitter) {
	songId := value
	song, ok := mr.db.GetSong(nil, songId)
	if !ok {
		log.Printf("song %s does not exist", songId)
		return
	}

	if song.Meta.Artist == "" || song.Info.State != "active" {
		return
	}

	// Just emit everything under a single key so we're running a single
	// Reduce() call.
	kv := mr.proto.MarshalKV("songs", songId)
	emitter.Emit(kv.Key, kv.Value)
}

func (mr *MRVerify) MapFinal(emitter dmrgo.Emitter) {
}

// Reduce collects statistics into a counter
func (mr *MRVerify) Reduce(key string, values []string, emitter dmrgo.Emitter) {
	var songIds []string
	mr.proto.UnmarshalKVs(key, values, &key, &songIds)

	added := 0
	removed := 0

	// Read current index contents.
	idxContents := sliceToMap(mr.idx.Scan())

	// Compute the set difference.
	for _, songId := range songIds {
		_, ok := idxContents[songId]
		if ok {
			delete(idxContents, songId)
		}
		// If 'rebuild' is set, add the document anyway.
		if *rebuild || !ok {
			log.Printf("add: %s", songId)
			mr.addSong(songId)
			added++
		}
	}
	for songId, _ := range idxContents {
		log.Printf("delete: %s", songId)
		mr.deleteSong(songId)
		removed++
	}

	log.Printf("added: %d, removed: %d", added, removed)
}

func main() {
	flag.Parse()

	if err := config.Read(*configFile); err != nil {
		log.Fatalf("Error: could not read config file %s: %s", *configFile, err.Error())
	}

	dbImpl := database.NewDatabaseImplFromConfig(config)
	idx := index.NewIndexClientFromConfig(config)
	mr := &MRVerify{
		db:    dbImpl,
		idx:   idx,
		proto: &dmrgo.JSONProtocol{},
	}
	dmrgo.Main(mr)
}
