package main

import (
	"flag"
	"fmt"
	"log"
	"time"

	"git.autistici.org/ale/djrandom/api"
	"git.autistici.org/ale/djrandom/services"
	"git.autistici.org/ale/djrandom/services/database"
	"git.autistici.org/ale/djrandom/util"
)

var (
	config     = util.NewConfig()
	configFile = flag.String("config", "/etc/djrandom/server.conf", "Config file location")
)

func createNewUser(db services.Database, username string, pw string) {
	user := &api.User{
		Email:       username,
		CreatedAt:   time.Now().Unix(),
		InvitedBy:   "admin",
		InvitesLeft: 3,
		State:       "active",
	}
	user.SetPassword(pw)
	err := db.PutUser(nil, user)
	if err != nil {
		log.Fatalf("Error creating user: %s", err)
	}
}

func setPassword(db services.Database, username string, pw string) {
	s, err := db.NewSession()
	if err != nil {
		log.Fatalf("Error connecting to db: %s", err)
	}
	defer s.Close()

	user, ok := db.GetUser(s, username)
	if !ok {
		log.Fatalf("User does not exist")
	}
	user.SetPassword(pw)
	err = db.PutUser(s, user)
	if err != nil {
		log.Fatalf("Error updating user: %s", err)
	}
}

func createNewAuthKey(db services.Database, username string) {
	s, err := db.NewSession()
	if err != nil {
		log.Fatalf("Error connecting to db: %s", err)
	}
	defer s.Close()

	user, ok := db.GetUser(s, username)
	if !ok {
		log.Fatalf("User does not exist")
	}
	authKey := user.NewAuthKey()
	err = db.PutAuthKey(s, authKey)
	if err != nil {
		log.Fatalf("Error saving auth key: %s", err)
	}
	fmt.Printf("auth_key=%s\nauth_secret=%s\n", authKey.KeyId, authKey.Secret)
}

func main() {
	flag.Parse()

	if err := config.Read(*configFile); err != nil {
		log.Fatalf("Error: could not read config file %s: %s", *configFile, err.Error())
	}

	db := database.NewDatabaseImplFromConfig(config)

	cmd := flag.Arg(0)
	switch cmd {
	case "add-user":
		createNewUser(db, flag.Arg(1), flag.Arg(2))
	case "gen-auth-key":
		createNewAuthKey(db, flag.Arg(1))
	case "set-password":
		setPassword(db, flag.Arg(1), flag.Arg(2))
	default:
		log.Fatalf("Unknown command '%s'", cmd)
	}
}
