package backend_sql

import (
	"errors"
	"fmt"
	"log"
	"strings"
	// "sync"
	"database/sql"

	// _ "code.google.com/p/gosqlite/sqlite"
	_ "github.com/mattn/go-sqlite3"

	"git.autistici.org/ale/djrandom/services/database"
)

type SqlBackend struct {
	db      *sql.DB
	getStmt *sql.Stmt
	putStmt *sql.Stmt
	delStmt *sql.Stmt
	// lock sync.Mutex
}

func mustPrepare(db *sql.DB, s string) *sql.Stmt {
	stmt, err := db.Prepare(s)
	if err != nil {
		log.Fatal("Error creating prepared statement: %s", err)
	}
	return stmt
}

func NewSqlBackend(path string) *SqlBackend {
	db, err := sql.Open("sqlite3", path)
	if err != nil {
		log.Fatal("Could not create database %s: %s", path, err.Error())
	}

	// Ignore errors from this statement.
	db.Exec("CREATE TABLE IF NOT EXISTS kv (key TEXT PRIMARY KEY NOT NULL, value BLOB);")

	return &SqlBackend{
		db:      db,
		getStmt: mustPrepare(db, "SELECT value FROM kv WHERE key = ? LIMIT 1"),
		delStmt: mustPrepare(db, "DELETE FROM kv WHERE key = ?"),
		putStmt: mustPrepare(db, "INSERT OR REPLACE INTO kv (key, value) VALUES (?, ?)"),
	}
}

func (be *SqlBackend) Close() {
	be.db.Close()
}

func (be *SqlBackend) makeRowId(bucket, key string) string {
	return fmt.Sprintf("%s/%s", bucket, key)
}

func (be *SqlBackend) parseRowId(row string) (string, string) {
	kvparts := strings.SplitN(row, "/", 2)
	return kvparts[0], kvparts[1]
}

func (be *SqlBackend) Get(bucket, key string) ([]byte, error) {
	row := be.makeRowId(bucket, key)

	tx, err := be.db.Begin()
	if err != nil {
		log.Printf("Get(%s): Error creating transaction: %s", row, err)
		return nil, err
	}
	defer tx.Rollback()

	s := tx.Stmt(be.getStmt)
	defer s.Close()

	var buf []byte
	err = s.QueryRow(row).Scan(&buf)
	if err != nil {
		if err != sql.ErrNoRows {
			log.Printf("Get(%s): Error selecting row: %s", row, err)
		}
		return nil, err
	}

	// 'buf' is not persistent, return a copy.
	data := make([]byte, len(buf))
	copy(data, buf)
	return data, nil
}

func (be *SqlBackend) Put(bucket, key string, data []byte) error {
	row := be.makeRowId(bucket, key)

	tx, err := be.db.Begin()
	if err != nil {
		log.Printf("Put(%s): Error creating transaction: %s", row, err)
		return err
	}

	s := tx.Stmt(be.putStmt)
	_, err = s.Exec(row, data)
	s.Close()
	if err != nil {
		tx.Rollback()
		return err
	}

	return tx.Commit()
}

func (be *SqlBackend) Del(bucket, key string) error {
	row := be.makeRowId(bucket, key)

	tx, err := be.db.Begin()
	if err != nil {
		log.Printf("Del(%s): Error creating transaction: %s", row, err)
		return err
	}

	s := tx.Stmt(be.delStmt)
	_, err = s.Exec(row)
	s.Close()
	if err != nil {
		tx.Rollback()
		return err
	}

	return tx.Commit()
}

type sqliteIteratorKeyPair struct {
	key   string
	value []byte
}

type SqliteIterator struct {
	C chan *sqliteIteratorKeyPair
}

var (
	EOF = errors.New("EOF")
)

func (i *SqliteIterator) Next() (string, []byte, error) {
	pair := <-i.C
	if pair == nil {
		return "", nil, EOF
	}
	return pair.key, pair.value, nil
}

func (i *SqliteIterator) Close() {
}

func (be *SqlBackend) Scan(bucket, startKey, endKey string, limit int) (database.BackendIterator, error) {
	i := &SqliteIterator{
		C: make(chan *sqliteIteratorKeyPair),
	}

	go be.scanInner(bucket, startKey, endKey, limit, i.C)
	return i, nil
}

func (be *SqlBackend) scanInner(bucket, startKey, endKey string, limit int, c chan *sqliteIteratorKeyPair) {
	tx, err := be.db.Begin()
	if err != nil {
		log.Printf("Scan(): Error creating transaction: %s", err)
		return
	}
	defer tx.Rollback()

	sql := "SELECT key, value FROM kv WHERE key >= ? AND key < ? ORDER BY key ASC"
	if limit > 0 {
		sql = fmt.Sprintf("%s LIMIT %d", sql, limit)
	}
	startRow := be.makeRowId(bucket, startKey)
	endRow := be.makeRowId(bucket, endKey)

	s, err := tx.Prepare(sql)
	if err != nil {
		return
	}
	defer s.Close()

	rows, err := s.Query(startRow, endRow)
	if err != nil {
		return
	}

	for rows.Next() {
		var row string
		var value []byte
		if err := rows.Scan(&row, &value); err != nil {
			log.Printf("NamedScan() error: %s", err)
			return
		}

		// 'value' is not persistent, make a copy.
		buf := make([]byte, len(value))
		copy(buf, value)
		_, key := be.parseRowId(row)
		c <- &sqliteIteratorKeyPair{key: key, value: buf}
	}
	err = rows.Err()
	if err != nil {
		log.Printf("Rows() error: %s", err)
	}
	close(c)
}
