package backend_sqlite

import (
	"errors"
	"fmt"
	"log"
	"strings"
	// "sync"

	sqlite "github.com/gwenn/gosqlite"

	"git.autistici.org/ale/djrandom/services/database"
)

var NotFound = errors.New("Not Found")

type SqliteBackend struct {
	conn *sqlite.Conn
	// lock sync.Mutex
}

func NewSqliteBackend(path string) *SqliteBackend {
	db, err := sqlite.Open(path)
	if err != nil {
		log.Fatal("Could not create database %s: %s", path, err.Error())
	}

	// Ignore errors from this statement.
	db.Exec("CREATE TABLE IF NOT EXISTS kv (key TEXT PRIMARY KEY NOT NULL, value BLOB);")

	return &SqliteBackend{conn: db}
}

func (db *SqliteBackend) Close() {
	db.conn.Close()
}

func (db *SqliteBackend) makeRowId(bucket, key string) string {
	return fmt.Sprintf("%s/%s", bucket, key)
}

func (db *SqliteBackend) parseRowId(row string) (string, string) {
	kvparts := strings.SplitN(row, "/", 2)
	return kvparts[0], kvparts[1]
}

func (db *SqliteBackend) Get(bucket, key string) ([]byte, error) {
	// db.lock.Lock()
	// defer db.lock.Unlock()
	row := db.makeRowId(bucket, key)

	s, err := db.conn.Prepare("SELECT value FROM kv WHERE key = ? LIMIT 1", row)
	if err != nil {
		log.Printf("Get(%s): Error getting row: %s", row, err.Error())
		return nil, err
	}
	defer s.Finalize()

	var buf []byte
	ok, err := s.SelectOneRow(&buf)
	if err != nil {
		log.Printf("Get(%s): Error selecting row: %s", row, err.Error())
		return nil, err
	}
	if !ok {
		return nil, NotFound
	}

	// 'buf' is not persistent, return a copy.
	data := make([]byte, len(buf))
	copy(data, buf)
	return data, nil
}

func (db *SqliteBackend) Put(bucket, key string, data []byte) error {
	// db.lock.Lock()
	// defer db.lock.Unlock()
	row := db.makeRowId(bucket, key)

	s, err := db.conn.Prepare("INSERT OR REPLACE INTO kv (key, value) VALUES (?, ?)")
	if err != nil {
		return err
	}
	defer s.Finalize()
	_, err = s.Insert(row, data)
	return err
}

func (db *SqliteBackend) Del(bucket, key string) error {
	row := db.makeRowId(bucket, key)

	s, err := db.conn.Prepare("DELETE FROM kv WHERE key = ?")
	if err != nil {
		return err
	}
	defer s.Finalize()
	return s.Exec(row)
}

type sqliteIteratorKeyPair struct {
	key   string
	value []byte
}

type SqliteIterator struct {
	C chan *sqliteIteratorKeyPair
}

var (
	EOF = errors.New("EOF")
)

func (i *SqliteIterator) Next() (string, []byte, error) {
	pair := <-i.C
	if pair == nil {
		return "", nil, EOF
	}
	return pair.key, pair.value, nil
}

func (i *SqliteIterator) Close() {
}

func (db *SqliteBackend) Scan(bucket, startKey, endKey string, limit int, keysOnly bool) (database.BackendIterator, error) {
	i := &SqliteIterator{
		C: make(chan *sqliteIteratorKeyPair),
	}

	var sql string
	if keysOnly {
		sql = "SELECT key FROM kv WHERE key >= ? AND key < ? ORDER BY key ASC"
	} else {
		sql = "SELECT key, value FROM kv WHERE key >= ? AND key < ? ORDER BY key ASC"
	}
	if limit > 0 {
		sql = fmt.Sprintf("%s LIMIT %d", sql, limit)
	}
	startRow := db.makeRowId(bucket, startKey)
	endRow := db.makeRowId(bucket, endKey)

	s, err := db.conn.Prepare(sql, startRow, endRow)
	if err != nil {
		return nil, err
	}

	go func() {
		defer s.Finalize()

		err = s.Select(func(s *sqlite.Stmt) error {
			var row string
			var value []byte

			var err error
			if keysOnly {
				err = s.NamedScan("key", &row, "value", &value)
			} else {
				err = s.NamedScan("key", &row)
			}
			if err != nil {
				log.Printf("NamedScan() error: %s", err.Error())
				return err
			}

			// 'value' is not persistent, make a copy.
			if value != nil {
				buf := make([]byte, len(value))
				copy(buf, value)
				value = buf
			}

			_, key := db.parseRowId(row)
			i.C <- &sqliteIteratorKeyPair{key: key, value: value}
			return nil
		})
		if err != nil {
			log.Printf("Scan() error: %s", err.Error())
		}
		close(i.C)
	}()

	return i, nil
}
