package main

import (
	"encoding/json"
	"flag"
	"log"
	"os"

	"git.autistici.org/ale/djrandom/services/database"
	"git.autistici.org/ale/djrandom/services/database/backend_leveldb"
	"git.autistici.org/ale/djrandom/services/database/backend_sqlite"
	"git.autistici.org/ale/djrandom/util"
)

var (
	config     = util.NewConfig()
	configFile = flag.String("config", "/etc/djrandom/server.conf", "Config file location")
	dump       = flag.Bool("dump", false, "Dump the database")
	restore    = flag.Bool("restore", false, "Restore a database dump")
	backend    = flag.String("backend", "sqlite", "Backend type (leveldb/sqlite)")
)

var allBuckets = []string{
	"song",
	"audio",
	"fp",
	"user",
	"auth_key",
	"playlog",
	"markov",
}

func doDump(db database.Backend) {
	output := make(map[string]map[string][]byte)
	for _, bucket := range allBuckets {
		iter, err := db.Scan(bucket, "", "\xff", -1, false)
		if err != nil {
			log.Fatalf("Scan error: %s", err)
		}

		bucketDump := make(map[string][]byte)
		for {
			key, data, err := iter.Next()
			if err != nil {
				break
			}
			bucketDump[key] = data
		}
		output[bucket] = bucketDump
	}

	json.NewEncoder(os.Stdout).Encode(output)
}

func doRestore(db database.Backend) {
	var input map[string]map[string][]byte
	if err := json.NewDecoder(os.Stdin).Decode(&input); err != nil {
		log.Fatal("Error decoding input JSON: %s", err)
	}

	for bucket, bucketDump := range input {
		for key, value := range bucketDump {
			if err := db.Put(bucket, key, value); err != nil {
				log.Fatal("Put(%s, %s) failed: %s", bucket, key, err)
			}
		}
	}
}

func main() {
	flag.Parse()

	if err := config.Read(*configFile); err != nil {
		log.Fatalf("Error: could not read config file %s: %s", *configFile, err.Error())
	}

	if !*dump && !*restore {
		log.Fatal("You need to specify either --dump or --restore")
	}

	dbPath := config.MustGetString("db_path")
	var db database.Backend
	switch *backend {
	case "leveldb":
		db = backend_leveldb.NewLevelDbBackend(dbPath)
	case "sqlite":
		db = backend_sqlite.NewSqliteBackend(dbPath)
	default:
		log.Fatalf("Unknown backend '%s'", *backend)
	}

	if *dump {
		doDump(db)
	} else {
		doRestore(db)
	}
}
