package database

import (
	"encoding/json"
	"errors"
	"log"
	"net/rpc"

	"git.autistici.org/ale/djrandom/services"
)

var (
	EOF = errors.New("EOF")
)

type ClientCodec interface {
	Marshal(interface{}) ([]byte, error)
	Unmarshal([]byte, interface{}) error
}

type jsonCodec struct{}

func (j jsonCodec) Marshal(obj interface{}) ([]byte, error) {
	return json.Marshal(obj)
}

func (j jsonCodec) Unmarshal(data []byte, obj interface{}) error {
	return json.Unmarshal(data, obj)
}

type DatabaseRpcClient struct {
	codec    ClientCodec
	endpoint string
}

type DatabaseRpcClientSession struct {
	client *rpc.Client
	codec  ClientCodec
}

func NewDatabaseRpcClient(endpoint string) *DatabaseRpcClient {
	return &DatabaseRpcClient{
		codec:    jsonCodec{},
		endpoint: endpoint,
	}
}

func (c *DatabaseRpcClient) NewSession() (services.Session, error) {
	client, err := rpc.DialHTTP("tcp", c.endpoint)
	if err != nil {
		log.Printf("Database: error connecting to %s: %s", c.endpoint, err)
		return nil, err
	}
	return &DatabaseRpcClientSession{client, c.codec}, nil
}

// Session methods.

func (s *DatabaseRpcClientSession) Get(bucket, key string, obj interface{}) error {
	req := &DbGetRequest{
		Bucket: bucket,
		Key:    key,
	}
	var resp DbGetResponse
	if err := s.client.Call("Db.Get", req, &resp); err != nil {
		return err
	}
	return s.codec.Unmarshal(resp.Data, obj)
}

func (s *DatabaseRpcClientSession) Put(bucket, key string, obj interface{}) error {
	data, err := s.codec.Marshal(obj)
	if err != nil {
		return err
	}
	req := &DbPutRequest{
		Bucket: bucket,
		Key:    key,
		Data:   data,
	}
	var resp EmptyResponse
	return s.client.Call("Db.Put", req, &resp)
}

func (s *DatabaseRpcClientSession) Del(bucket, key string) error {
	req := &DbDelRequest{
		Bucket: bucket,
		Key:    key,
	}
	var resp EmptyResponse
	return s.client.Call("Db.Del", req, &resp)
}

type resultIterator struct {
	results []Datum
	codec   ClientCodec
	i       int
}

func (r *resultIterator) Next(obj interface{}) (string, error) {
	if r.i >= len(r.results) {
		return "", EOF
	}
	if obj != nil {
		if err := r.codec.Unmarshal(r.results[r.i].Data, obj); err != nil {
			return "", err
		}
	}
	key := r.results[r.i].Key
	r.i++
	return key, nil
}

func (s *DatabaseRpcClientSession) doScan(bucket, startKey, endKey string, limit int, keysOnly bool) (services.Iterator, error) {
	req := &DbScanRequest{
		Bucket:   bucket,
		StartKey: startKey,
		EndKey:   endKey,
		Limit:    limit,
		KeysOnly: keysOnly,
	}
	var resp DbScanResponse
	if err := s.client.Call("Db.Scan", req, &resp); err != nil {
		return nil, err
	}

	iter := &resultIterator{
		results: resp.Results,
		codec:   s.codec,
	}
	return iter, nil
}

func (s *DatabaseRpcClientSession) Scan(bucket, startKey, endKey string, limit int) (services.Iterator, error) {
	return s.doScan(bucket, startKey, endKey, limit, false)
}

func (s *DatabaseRpcClientSession) ScanKeys(bucket, startKey, endKey string, limit int) (services.Iterator, error) {
	return s.doScan(bucket, startKey, endKey, limit, true)
}

func (s *DatabaseRpcClientSession) Close() {
	s.client.Close()
}

/**

var (
	maxRetries = 10
	retryDelay = 100 * time.Millisecond
)

func (c *DatabaseRpcClient) connect() (*rpc.Client, error) {
	c.clientLock.Lock()
	defer c.clientLock.Unlock()

	if c.client == nil {
		var err error
		c.client, err = rpc.DialHTTP("tcp", c.endpoint)
		if err != nil {
			log.Printf("Database: error connecting to %s: %s", c.endpoint, err)
			return nil, err
		}
	}
	return c.client, nil
}

func (c *DatabaseRpcClient) invalidateConnection() {
	c.clientLock.Lock()
	defer c.clientLock.Unlock()
	c.client = nil
}

func (c *DatabaseRpcClient) retryCall(method string, req interface{}, resp interface{}) error {
	var lastErr error
	for i := 0; i < maxRetries; i++ {
		client, err := c.connect()
		if err != nil {
			lastErr = err
		} else {
			lastErr := client.Call(method, req, resp)
			if lastErr == nil {
				return lastErr
			}
			log.Printf("Database: error calling %s: %s", method, lastErr)
			client.Close()
			c.invalidateConnection()
		}
		time.Sleep(retryDelay)
	}
	return lastErr
}
**/
