// The database service consists of a set of high-level operations
// built on top of a simple key/value store. 
//
// The backend implementation is abstracted in the DatabaseClient
// interface, and is currently implemented as a single-instance RPC
// server using an SQLite backend.  The implementation can be replaced
// by any key/value store that supports buckets and range scans.
//
// The high-level API implements some simple relational behavior for
// the more complex objects (such as Song), and it hides the bucket
// names from the caller. Access is managed via Session objects, which
// are supposed to be short-lived and expose the low-level API. One
// day Sessions might acquire transactional semantics, but it
// currently has none.
//
package database

import (
	"fmt"
	"log"
	"math/rand"
	"strings"
	"sync"
	"time"

	"git.autistici.org/ale/djrandom/api"
	"git.autistici.org/ale/djrandom/services"
	"git.autistici.org/ale/djrandom/util"
)

const (
	// Key prefixes for various object types.
	songBucket        = "song"
	audioFileBucket   = "audio"
	fingerprintBucket = "fp"
	userBucket        = "user"
	authKeyBucket     = "auth_key"
	playlogBucket     = "playlog"
	markovBucket      = "markov"
)

type DatabaseImpl struct {
	client services.DatabaseClient
}

func NewDatabaseImpl(client services.DatabaseClient) *DatabaseImpl {
	return &DatabaseImpl{client: client}
}

func NewDatabaseImplFromConfig(config *util.Config) *DatabaseImpl {
	client := NewDatabaseRpcClient(config.MustGetString("db_addr"))
	return NewDatabaseImpl(client)
}

func (db *DatabaseImpl) NewSession() (services.Session, error) {
	return db.client.NewSession()
}

func (db *DatabaseImpl) uniqueToken(base string) string {
	return fmt.Sprintf("%s:%08x:%016x",
		base,
		time.Now().Unix(),
		rand.Int63())
}

func (db *DatabaseImpl) makeKey(parts ...string) string {
	return strings.Join(parts, "/")
}

func (db *DatabaseImpl) GetSong(s services.Session, key string) (*api.Song, bool) {
	if s == nil {
		var serr error
		if s, serr = db.client.NewSession(); serr != nil {
			return nil, false
		}
		defer s.Close()
	}

	var songIntl api.SongIntl
	err := s.Get(songBucket, key, &songIntl)
	if err != nil {
		log.Printf("GetSong(%s): error %s", key, err)
		return nil, false
	}

	song := api.Song{
		Id:    songIntl.Id,
		Meta:  songIntl.Meta,
		Info:  songIntl.Info,
		Files: make([]*api.AudioFile, len(songIntl.Files)),
	}

	var wg sync.WaitGroup

	for i, audioHash := range songIntl.Files {
		wg.Add(1)
		go func(audioHash string, idx int) {
			af, ok := db.GetAudioFile(s, audioHash)
			if !ok {
				log.Printf("GetSong(%s): AudioFile not found: %s", key, audioHash)
			} else {
				song.Files[idx] = af
			}
			wg.Done()
		}(audioHash, i)
	}

	wg.Wait()

	return &song, true
}

func (db *DatabaseImpl) PutSong(s services.Session, song *api.Song) error {
	if s == nil {
		var serr error
		if s, serr = db.client.NewSession(); serr != nil {
			return serr
		}
		defer s.Close()
	}

	songIntl := api.SongIntl{
		Id:    song.Id,
		Meta:  song.Meta,
		Info:  song.Info,
		Files: make([]string, len(song.Files)),
	}

	var wg sync.WaitGroup

	for i, af := range song.Files {
		songIntl.Files[i] = af.MD5
		wg.Add(1)
		go func(af *api.AudioFile) {
			err := db.PutAudioFile(s, af)
			if err != nil {
				log.Printf("PutSong(%s): Error saving audioFile %s: %s", song.Id, af.MD5, err.Error())
			}
			wg.Done()
		}(af)
	}

	wg.Wait()

	return s.Put(songBucket, song.Id, songIntl)
}

func (db *DatabaseImpl) GetAudioFile(s services.Session, key string) (*api.AudioFile, bool) {
	var af api.AudioFile
	if err := db.Get(s, audioFileBucket, key, &af); err != nil {
		return nil, false
	}
	return &af, true
}

func (db *DatabaseImpl) PutAudioFile(s services.Session, af *api.AudioFile) error {
	return db.Put(s, audioFileBucket, af.MD5, af)
}

func (db *DatabaseImpl) GetUser(s services.Session, email string) (*api.User, bool) {
	var user api.User
	if err := db.Get(s, userBucket, email, &user); err != nil {
		return nil, false
	}
	return &user, true
}

func (db *DatabaseImpl) PutUser(s services.Session, user *api.User) error {
	return db.Put(s, userBucket, user.Email, user)
}

func (db *DatabaseImpl) GetAuthKey(s services.Session, keyId string) (*api.AuthKey, bool) {
	var authKey api.AuthKey
	if err := db.Get(s, authKeyBucket, keyId, &authKey); err != nil {
		return nil, false
	}
	return &authKey, true
}

func (db *DatabaseImpl) PutAuthKey(s services.Session, authKey *api.AuthKey) error {
	return db.Put(s, authKeyBucket, authKey.KeyId, authKey)
}

func (db *DatabaseImpl) GetSongFingerprint(s services.Session, key string) (*api.Fingerprint, error) {
	var fp api.Fingerprint
	err := db.Get(s, fingerprintBucket, key, &fp)
	return &fp, err
}

func (db *DatabaseImpl) PutSongFingerprint(s services.Session, key string, fp *api.Fingerprint) error {
	return db.Put(s, fingerprintBucket, key, fp)
}

func (db *DatabaseImpl) GetPlayLog(s services.Session, key string) (*api.PlayLogEntry, bool) {
	var pl api.PlayLogEntry
	if err := db.Get(s, playlogBucket, key, &pl); err != nil {
		return nil, false
	}
	return &pl, true
}

func (db *DatabaseImpl) AppendPlayLog(s services.Session, entry *api.PlayLogEntry) error {
	key := fmt.Sprintf("%s/%s", entry.User, db.uniqueToken("pl"))
	return db.Put(s, playlogBucket, key, entry)
}

func (db *DatabaseImpl) GetMarkovMap(s services.Session, user string) (*api.MarkovMap, bool) {
	var m api.MarkovMap
	if err := db.Get(s, markovBucket, user, &m); err != nil {
		return nil, false
	}
	return &m, true
}

func (db *DatabaseImpl) PutMarkovMap(s services.Session, user string, m *api.MarkovMap) error {
	return db.Put(s, markovBucket, user, m)
}

// Get exposes the generic 'get' method with optional session wrap.
func (db *DatabaseImpl) Get(s services.Session, bucket string, key string, obj interface{}) error {
	if s == nil {
		var serr error
		if s, serr = db.client.NewSession(); serr != nil {
			return serr
		}
		defer s.Close()
	}
	return s.Get(bucket, key, obj)
}

// Put exposes the generic 'put' method with optional session wrap.
func (db *DatabaseImpl) Put(s services.Session, bucket string, key string, obj interface{}) error {
	if s == nil {
		var serr error
		if s, serr = db.client.NewSession(); serr != nil {
			return serr
		}
		defer s.Close()
	}
	return s.Put(bucket, key, obj)
}

// Del exposes the generic 'del' method with optional session wrap.
func (db *DatabaseImpl) Del(s services.Session, bucket string, key string) error {
	if s == nil {
		var serr error
		if s, serr = db.client.NewSession(); serr != nil {
			return serr
		}
		defer s.Close()
	}
	return s.Del(bucket, key)
}
