package database

import (
	"encoding/json"
	"testing"
	"time"

	"git.autistici.org/ale/djrandom/api"
	"git.autistici.org/ale/djrandom/services"
	"git.autistici.org/ale/djrandom/services/database/backend_sqlite"
)

func createSong() *api.Song {
	af := api.AudioFile{
		MD5:        "aaabbbccc",
		Size:       1234,
		SampleRate: 44100,
		Channels:   2,
		Duration:   1.12,
		SongId:     "song1",
	}
	return &api.Song{
		Id:    "song1",
		Files: []*api.AudioFile{&af},
		Meta: api.Meta{
			Artist: "artist",
			Title:  "title",
			Album:  "an album",
			Year:   1970,
		},
		Info: api.SongInfo{
			State:     "ok",
			CreatedAt: time.Now().Unix(),
		},
	}
}

// Short-circuit the rpc client with a local Sqlite database.
// Implements the DatabaseClient interface.
type testLocalDatabaseClient struct {
	be *backend_sqlite.SqliteBackend
}

// Session for the above.
type testLocalDatabaseSession struct {
	be *backend_sqlite.SqliteBackend
}

func newTestLocalDatabaseClient() *testLocalDatabaseClient {
	return &testLocalDatabaseClient{
		backend_sqlite.NewSqliteBackend(":memory:"),
	}
}

func (db *testLocalDatabaseClient) Close() {
	db.be.Close()
	// os.Remove(db.path)
}

func (db *testLocalDatabaseClient) NewSession() (services.Session, error) {
	return &testLocalDatabaseSession{db.be}, nil
}

func (db *testLocalDatabaseSession) Close() {}

func (db *testLocalDatabaseSession) Get(bucket, key string, obj interface{}) error {
	data, err := db.be.Get(bucket, key)
	if err != nil {
		return err
	}
	return json.Unmarshal(data, obj)
}

func (db *testLocalDatabaseSession) Put(bucket, key string, obj interface{}) error {
	data, err := json.Marshal(obj)
	if err != nil {
		return err
	}
	return db.be.Put(bucket, key, data)
}

func (db *testLocalDatabaseSession) Del(bucket, key string) error {
	return db.be.Del(bucket, key)
}

type testResultIterator struct {
	i BackendIterator
}

func (r testResultIterator) Next(obj interface{}) (string, error) {
	key, data, err := r.i.Next()
	if err != nil {
		return key, err
	}
	return key, json.Unmarshal(data, obj)
}

func (db *testLocalDatabaseSession) Scan(bucket, startKey, endKey string, limit int) (services.Iterator, error) {
	i, err := db.be.Scan(bucket, startKey, endKey, limit)
	if err != nil {
		return nil, err
	}
	return &testResultIterator{i}, nil
}

func TestDatabase_PutSong(t *testing.T) {
	client := newTestLocalDatabaseClient()
	defer client.Close()
	db := NewDatabaseImpl(client)

	song := createSong()
	err := db.PutSong(nil, song)
	if err != nil {
		t.Fatalf("PutSong() failed: %s", err.Error())
	}
}

func TestDatabase_ModifySong(t *testing.T) {
	client := newTestLocalDatabaseClient()
	defer client.Close()
	db := NewDatabaseImpl(client)

	song := createSong()
	err := db.PutSong(nil, song)
	if err != nil {
		t.Fatalf("PutSong() failed: %s", err.Error())
	}

	song, ok := db.GetSong(nil, "song1")
	if !ok {
		t.Fatalf("GetSong() failed, not found")
	}

	// Add a couple of new AudioFiles to the song (it should be
	// autosaved to the db).
	newFiles := []*api.AudioFile{}
	newFiles = append(newFiles, &api.AudioFile{
		MD5:        "cccdddeee",
		Size:       2345,
		SampleRate: 44100,
		Channels:   2,
		Duration:   2.00,
		SongId:     "song1",
	})
	newFiles = append(newFiles, &api.AudioFile{
		MD5:        "dddeeefff",
		Size:       3456,
		SampleRate: 44100,
		Channels:   2,
		Duration:   3.00,
		SongId:     "song1",
	})

	song.Files = append(song.Files, newFiles...)

	err = db.PutSong(nil, song)
	if err != nil {
		t.Fatalf("PutSong() / 2 failed: %s", err.Error())
	}

	// Check that we can retrieve the new AudioFile.
	af2, ok := db.GetAudioFile(nil, "cccdddeee")
	if !ok {
		t.Fatalf("GetAudioFile() failed for new AudioFile")
	}
	if af2.Channels != 2 {
		t.Fatalf("Result of GetAudioFile() does not match")
	}
}
