package index

import (
	"log"
	"net/rpc"
	"sync"

	"git.autistici.org/ale/djrandom/api"
	"git.autistici.org/ale/djrandom/partition"
	"git.autistici.org/ale/djrandom/util"
)

// IndexClient is the distributed index client stub.
type IndexClient struct {
	P *partition.PartitionMap
}

func NewIndexClient(pmap *partition.PartitionMap) *IndexClient {
	return &IndexClient{P: pmap}
}

func NewIndexClientFromConfig(config *util.Config) *IndexClient {
	pmap := partition.MustLoadPartitionMap(
		config.MustGetString("index_pmap"))
	pmap.HandleDebug("index")
	return NewIndexClient(pmap)
}

// AddDoc adds a Document to the index.
func (ic *IndexClient) AddDoc(docId string, doc api.Document) error {
	target := ic.P.GetTarget(docId)
	client, err := rpc.DialHTTP("tcp", target)
	if err != nil {
		return err
	}
	defer client.Close()

	ar := &AddDocRequest{
		DocId: docId,
		Doc:   doc,
	}
	var resp bool
	err = client.Call("IndexService.AddDoc", ar, &resp)
	if err != nil {
		return err
	}

	return nil
}

// DeleteDoc removes a document from the index, given its ID.
func (ic *IndexClient) DeleteDoc(docId string) error {
	target := ic.P.GetTarget(docId)
	client, err := rpc.DialHTTP("tcp", target)
	if err != nil {
		return err
	}
	defer client.Close()

	req := &DeleteDocRequest{DocId: docId}
	var resp DeleteDocResponse
	err = client.Call("IndexService.DeleteDoc", req, &resp)
	if err != nil {
		return err
	}

	return nil
}

// Search runs a query against the index and returns a set of document IDs.
func (ic *IndexClient) Search(query string) api.Set {
	result := make(api.Set)
	ch := make(chan api.Set)
	defer close(ch)
	var wg sync.WaitGroup

	for _, t := range ic.P.AllTargets() {
		wg.Add(1)
		go func(target string) {
			client, err := rpc.DialHTTP("tcp", target)
			if err != nil {
				log.Printf("Error in distributed search on target %s: %s", target, err)
				wg.Done()
				return
			}
			defer client.Close()

			req := &SearchRequest{Query: query}
			var resp SearchResponse
			err = client.Call("IndexService.Search", req, &resp)
			if err != nil {
				log.Printf("Remote error in distributed search on target %s: %s", target, err)
				wg.Done()
				return
			}
			ch <- resp.Results
			wg.Done()
			return
		}(t)
	}

	go func() {
		for partial := range ch {
			result.Update(partial)
		}
	}()

	wg.Wait()

	return result
}

// Scan returns a list of all the document IDs contained in the index
// (warning: potentially a very large list!).
func (ic *IndexClient) Scan() []string {
	ids := make([]string, 0)
	ch := make(chan []string)
	defer close(ch)
	var wg sync.WaitGroup

	for _, t := range ic.P.AllTargets() {
		wg.Add(1)
		go func(target string) {
			client, err := rpc.DialHTTP("tcp", target)
			if err != nil {
				log.Printf("Error in distributed scan on target %s: %s", target, err)
				wg.Done()
				return
			}
			defer client.Close()

			req := &ScanRequest{}
			var resp ScanResponse
			if err := client.Call("IndexService.Scan", req, &resp); err == nil {
				ch <- resp.DocIds
			}

			wg.Done()
		}(t)
	}

	go func() {
		for partial := range ch {
			ids = append(ids, partial...)
		}
	}()

	wg.Wait()

	return ids
}
