// The index package implements a distributed search index.
//
// It is basically a simple reverse document index, implemented as a
// PartitionedService, which keeps the entire database in-memory (and
// occasionally checkpoints to disk).
//
// The query engine is not particularly clever or efficient, and the
// query expression language is rudimental, but the implementation is
// very simple and straightforward. And, of course, the entire service
// can be replaced by anything that implements the Index interface.
//
package index

import (
	"encoding/gob"
	"log"
	"os"
	"sync"

	"git.autistici.org/ale/djrandom/api"
)

// Index is an index for a single attribute.  Each key can have
// multiple values.
type Index struct {
	Name   string
	Hits   int64
	Misses int64
	M      map[string]api.Set
	lock   sync.RWMutex
}

func NewIndex(name string) *Index {
	return &Index{
		Name:   name,
		Hits:   0,
		Misses: 0,
		M:      make(map[string]api.Set),
	}
}

// Add a key/value pair to the index.
func (i *Index) Add(key string, value string) {
	i.lock.Lock()
	defer i.lock.Unlock()

	set, ok := i.M[key]
	if !ok {
		set = make(api.Set)
		i.M[key] = set
	}
	set.Add(value)
}

// Discard a key from the index.
func (i *Index) Discard(key string, value string) {
	i.lock.Lock()
	defer i.lock.Unlock()

	set, ok := i.M[key]
	if ok {
		set.Discard(value)
	}
}

// Get returns all values for key.
func (i *Index) Get(key string) (set api.Set, ok bool) {
	i.lock.RLock()
	defer i.lock.RUnlock()

	set, ok = i.M[key]
	if ok {
		i.Hits++
	} else {
		i.Misses++
	}
	return
}

type MultiIndex struct {
	// The various indexes we built.
	Indexes map[string]*Index

	// The reverse docId -> values map which we're going to use to
	// remove documents from the index.
	D map[string]api.Document
}

func NewMultiIndex() *MultiIndex {
	return &MultiIndex{
		Indexes: make(map[string]*Index),
		D:       make(map[string]api.Document),
	}
}

func LoadMultiIndex(path string) (*MultiIndex, error) {
	file, err := os.Open(path)
	if err != nil {
		return nil, err
	}
	defer file.Close()
	var is MultiIndex
	if err = gob.NewDecoder(file).Decode(&is); err != nil {
		return nil, err
	}
	return &is, nil
}

func (is *MultiIndex) Save(path string) error {
	file, err := os.Create(path)
	if err != nil {
		return err
	}
	defer file.Close()
	return gob.NewEncoder(file).Encode(is)
}

// AddDoc adds a document to the index.
func (is *MultiIndex) AddDoc(docId string, doc api.Document) error {
	// Are we re-indexing this document? If so, clear the existing
	// values before proceeding.
	_, ok := is.D[docId]
	if ok {
		is.DeleteDoc(docId)
	}

	for indexName, values := range doc {
		idx, ok := is.Indexes[indexName]
		if !ok {
			idx = NewIndex(indexName)
			is.Indexes[indexName] = idx
		}
		for _, v := range values {
			idx.Add(v, docId)
		}
	}
	is.D[docId] = doc

	// This implementation always succeeds.
	return nil
}

// DeleteDoc removes a document from the index.
func (is *MultiIndex) DeleteDoc(docId string) {
	doc, ok := is.D[docId]
	if !ok {
		return
	}

	// Remove the values associated with this document from
	// all the indexes we maintain.
	for indexName, values := range doc {
		idx, ok := is.Indexes[indexName]
		if !ok {
			continue
		}
		for _, v := range values {
			idx.Discard(v, docId)
		}
	}
	delete(is.D, docId)
}

// Lookup performs a point lookup in the index, returns a list of
// document IDs.
func (is *MultiIndex) Lookup(key string, value string) api.Set {
	if idx, ok := is.Indexes[key]; ok {
		if result, ok := idx.Get(value); ok {
			return result
		}
	}
	return make(api.Set)
}

// Search parses and executes a query. It returns the list of matching
// document IDs.
func (is *MultiIndex) Search(query string) api.Set {

	// Parse the query into an AST.
	pquery := StringToQuery(query)
	if pquery == nil {
		return nil
	}

	// Execute the parsed query.
	log.Printf("Executing query: %s", pquery)
	return pquery.Eval(is)
}

// Scan outputs all document IDs to the given channel.
func (is *MultiIndex) Scan(out chan string) {
	for docId, _ := range is.D {
		out <- docId
	}
}
