package index

import (
	"fmt"
	_ "math/rand"
	"os"
	"testing"

	"git.autistici.org/ale/djrandom/api"
)

func TestIndex_Add(t *testing.T) {
	i := NewIndex("test")
	i.Add("key1", "value1")
	i.Add("key1", "value2")
	i.Add("key2", "value2")

	result, ok := i.Get("key1")
	if !ok {
		t.Fatalf("Get(key1) not found")
	}
	if len(result) != 2 {
		t.Fatalf("Unexpected number of results for Get(key1): %d", len(result))
	}
}

func TestMultiIndex_AddDoc(t *testing.T) {
	srv := NewMultiIndex()
	doc := make(api.Document)
	doc.Add("index1", "v1")
	doc.Add("index2", "v2")
	srv.AddDoc("doc1", doc)
}

func BenchmarkMultiIndex_AddDoc(b *testing.B) {
	srv := NewMultiIndex()
	for i := 0; i < b.N; i++ {
		value := fmt.Sprintf("%d", i)
		docId := fmt.Sprintf("doc%03d", i)
		doc := make(api.Document)
		doc.Add("index1", "all")
		doc.Add("index2", value)
		srv.AddDoc(docId, doc)
	}
}

func TestMultiIndex_Delete(t *testing.T) {
	srv := NewMultiIndex()
	doc1 := make(api.Document)
	doc1.Add("index1", "v1")
	srv.AddDoc("doc1", doc1)

	doc2 := make(api.Document)
	doc2.Add("index1", "v1")
	srv.AddDoc("doc2", doc2)

	srv.DeleteDoc("doc2")

	result := srv.Search("index1:v1")
	if len(result) != 1 {
		t.Fatalf("Unexpected result length: %d", len(result))
	}
}

func TestMultiIndex_Search(t *testing.T) {
	srv := NewMultiIndex()
	doc1 := make(api.Document)
	doc1.Add("index1", "v1")
	srv.AddDoc("doc1", doc1)

	doc2 := make(api.Document)
	doc2.Add("index1", "v1")
	srv.AddDoc("doc2", doc2)

	result := srv.Search("index1:v1")
	if len(result) != 2 {
		t.Fatalf("Wrong number of results: %d", len(result))
	}

	result = srv.Search("index1:novalue")
	if len(result) != 0 {
		t.Fatalf("Error querying for index with no value: found=%d", len(result))
	}

	result = srv.Search("noindex:v1")
	if len(result) != 0 {
		t.Fatalf("Error querying for missing index: found=%d", len(result))
	}
}

func TestMultiIndex_SearchMultiValuedDoc(t *testing.T) {
	srv := NewMultiIndex()
	doc1 := make(api.Document)
	doc1.Add("index1", "v1")
	doc1.Add("index1", "v2")
	srv.AddDoc("doc1", doc1)

	result := srv.Search("index1:v1")
	if len(result) != 1 {
		t.Fatalf("Wrong number of results for v1: %d", len(result))
	}

	result = srv.Search("index1:v2")
	if len(result) != 1 {
		t.Fatalf("Wrong number of results for v2: %d", len(result))
	}
}

func TestMultiIndex_SearchMultipleIndexes(t *testing.T) {
	srv := NewMultiIndex()

	doc := make(api.Document)
	doc.Add("index1", "v2")
	doc.Add("index2", "v1")
	doc.Add("index3", "v3")
	srv.AddDoc("doc3", doc)

	doc = make(api.Document)
	doc.Add("index1", "v2")
	doc.Add("index2", "v1")
	doc.Add("index3", "v4")
	srv.AddDoc("doc4", doc)

	doc = make(api.Document)
	doc.Add("index1", "v2")
	doc.Add("index2", "v3")
	doc.Add("index3", "v5")
	srv.AddDoc("doc5", doc)

	// Intersection of two result sets.
	result := srv.Search("index1:v2 index2:v1")
	if len(result) != 2 {
		t.Fatalf("Wrong number of results for 2-way AND merge: %d", len(result))
	}

	// Intersection of three result sets.
	result = srv.Search("index1:v2 index2:v1 index3:v4")
	if len(result) != 1 {
		t.Fatalf("Wrong number of results for 3-way AND merge: %d", len(result))
	}

	// Union of two result sets.
	// q := mapToSearchQuery(map[string]string{"index2": "v1", "index3": "v4"})
	// q.Op = api.SearchOpOR
	// result = srv.Search(q)
	// if len(result) != 2 {
	// 	t.Fatalf("Wrong number of results for 2-way OR merge: %d", len(result))
	// }
}

func TestMultiIndex_SaveAndRestore(t *testing.T) {
	srv := NewMultiIndex()
	doc1 := make(api.Document)
	doc1.Add("index1", "v1")
	doc1.Add("index1", "v2")
	srv.AddDoc("doc1", doc1)

	tmpf := ".indexsavetest"
	defer os.Remove(tmpf)

	if err := srv.Save(tmpf); err != nil {
		t.Fatalf("Save(): %s", err)
	}

	srv2, err := LoadMultiIndex(tmpf)
	if err != nil {
		t.Fatalf("Load(): %s", err)
	}

	result := srv2.Search("index1:v1")
	if len(result) != 1 {
		t.Fatalf("Index did not reload properly: %v", result)
	}
}
