package index

import (
	"log"
	"net/rpc"
	"time"

	"git.autistici.org/ale/djrandom/api"
	"git.autistici.org/ale/djrandom/partition"
)

type IndexService struct {
	partition.PartitionedServiceBase

	// The in-memory multi-index object.
	Index *MultiIndex

	// Client for rebalancing.
	Client *IndexClient

	// Quit channel for the background updater.
	quit chan bool
}

// Once in a while, dump the index state to disk.
func periodicSaveWorker(idx *MultiIndex, stateFile string, quit chan bool) {
	t := time.NewTicker(time.Second * 300)
	defer t.Stop()

	save := func() {
		if err := idx.Save(stateFile); err != nil {
			log.Printf("Error saving index: %s", err)
		}
	}

	for {
		select {
		case <-t.C:
			save()
		case <-quit:
			save()
			return
		}
	}
}

func NewIndexService(stateFile string, pmap *partition.PartitionMap, selfTarget string) *IndexService {
	idx, err := LoadMultiIndex(stateFile)
	if err != nil {
		log.Printf("Warning: could not read index from %s (%s), creating empty index", stateFile, err)
		idx = NewMultiIndex()
	} else {
		log.Printf("Loaded %d documents from %s", len(idx.D), stateFile)
	}

	quit := make(chan bool, 1)
	go periodicSaveWorker(idx, stateFile, quit)

	return &IndexService{
		partition.PartitionedServiceBase{
			Pmap:       pmap,
			SelfTarget: selfTarget,
		},
		idx,
		NewIndexClient(pmap),
		quit,
	}
}

func (is *IndexService) Stop() {
	is.quit <- true
}

func (is *IndexService) Scan() chan string {
	ch := make(chan string)

	go func() {
		is.Index.Scan(ch)
		close(ch)
	}()

	return ch
}

func (is *IndexService) Move(key string) error {
	doc, ok := is.Index.D[key]
	if !ok {
		return nil
	}

	err := is.Client.AddDoc(key, doc)
	if err != nil {
		return err
	}

	is.Index.DeleteDoc(key)

	return nil
}

// RPC Interface

// Just a wrapper that only exposes RPC methods.
type IndexServiceRPC struct {
	is *IndexService
}

type AddDocRequest struct {
	DocId string
	Doc   api.Document
}

type DeleteDocRequest struct {
	DocId string
}

type DeleteDocResponse struct{}

type SearchRequest struct {
	Query string
}

type SearchResponse struct {
	Results api.Set
}

type ScanRequest struct{}

type ScanResponse struct {
	DocIds []string
}

func (isrpc *IndexServiceRPC) AddDoc(req *AddDocRequest, resp *bool) error {
	*resp = true
	return isrpc.is.Index.AddDoc(req.DocId, req.Doc)
}

func (isrpc *IndexServiceRPC) DeleteDoc(req *DeleteDocRequest, resp *DeleteDocResponse) error {
	isrpc.is.Index.DeleteDoc(req.DocId)
	return nil
}

func (isrpc *IndexServiceRPC) Search(req *SearchRequest, resp *SearchResponse) error {
	resp.Results = isrpc.is.Index.Search(req.Query)
	return nil
}

func (isrpc *IndexServiceRPC) Scan(req *ScanRequest, resp *ScanResponse) error {
	// This should probably use isrpc.is.Scan()...  but we can
	// cheat this way and reduce memory allocations.
	ids := make([]string, len(isrpc.is.Index.D))
	i := 0
	for docId, _ := range isrpc.is.Index.D {
		ids[i] = docId
		i++
	}
	resp.DocIds = ids
	return nil
}

func RegisterService(is *IndexService) {
	rpc.RegisterName("IndexService", &IndexServiceRPC{is})
	rpc.HandleHTTP()
}
