// Server that implements the Storage API.

package main

import (
	"flag"
	"log"
	"net/rpc"
	"time"

	"git.autistici.org/ale/djrandom/api"
	"git.autistici.org/ale/djrandom/partition"
	"git.autistici.org/ale/djrandom/services/storage"
	"git.autistici.org/ale/djrandom/util"
	"git.autistici.org/ale/djrandom/util/daemon"
)

var (
	config     = util.NewConfig()
	configFile = flag.String("config", "/etc/djrandom/server.conf", "Config file location")
	addr       = flag.String("addr", ":3003", "TCP address to listen on.")
)

func main() {
	flag.Parse()

	if err := config.Read(*configFile); err != nil {
		log.Fatalf("Error: could not read config file %s: %s", *configFile, err.Error())
	}

	daemon.Setup()

	storagePmap := partition.MustLoadPartitionMap(
		config.MustGetString("storage_pmap"))
	storagePmap.HandleDebug("storage")
	storageAuthKey := &api.AuthKey{
		KeyId:  config.MustGetString("storage_key_id"),
		Secret: config.MustGetString("storage_key_secret"),
	}

	// Storage service.
	svc := storage.NewStorageService(
		config.MustGetString("storage_root"),
		storageAuthKey,
		storagePmap,
		config.MustGetString("storage_self"),
	)

	// Start background rebalancing worker.
	partition.RegisterStatsService(svc)
	go partition.RunRebalancer(svc, time.Second*900)

	storage.RegisterService(svc)

	// Start the RPC HTTP handler for debug functions.
	rpc.HandleHTTP()

	log.Printf("Starting storage server on %s", *addr)
	err := util.ListenAndServe(*addr, nil)
	if err != nil {
		log.Fatal("ListenAndServe: ", err)
	}
}
