package fingerprinting

import (
	// "bytes"
	"encoding/json"
	"errors"
	"fmt"
	"log"
	"net/http"

	"git.autistici.org/ale/djrandom/api"
)

var (
	echonestApiUrl = "http://developer.echonest.com/api/v4/song/identify"
)

type Echonest struct {
	endpoint string
}

func NewEchonest(apiKey string) *Echonest {
	return &Echonest{
		endpoint: fmt.Sprintf("%s?api_key=%s", echonestApiUrl, apiKey),
	}
}

// Identify performs a query to the Echonest public servers to
// identify a song given a fingerprint. The fingerprint is actually
// some JSON-encoded data coming directly from 'echoprint-codegen'.
func (e *Echonest) Identify(fingerprint string, version string) (*api.Meta, error) {
	url := fmt.Sprintf("%s&code=%s&version=%s", e.endpoint, fingerprint, version)
	resp, err := http.Get(url)
	if err != nil {
		log.Printf("Identify(): echonest query returned error: %s", err)
		return nil, err
	}
	defer resp.Body.Close()
	if resp.StatusCode != 200 {
		log.Printf("Identify(): echonest returned status %d: %v", resp.StatusCode, resp)
		return nil, err
	}

	var echonestResponse struct {
		Response struct {
			Status struct {
				Version string
				Message string
				Code    int
			}
			Songs []struct {
				ArtistName string `json:"artist_name"`
				Title      string `json:"title"`
				ArtistId   string `json:"artist_id"`
				Id         string `json:"id"`
				Message    string `json:"message"`
				Score      int    `json:"score"`
			}
		}
	}
	if err = json.NewDecoder(resp.Body).Decode(&echonestResponse); err != nil {
		log.Printf("Identify(): echonest returned bad response: %s", err)
		return nil, err
	}
	log.Printf("Identify(): response = %v", echonestResponse.Response)

	// Status.Code == 0 means ok, everything else is an error.
	if echonestResponse.Response.Status.Code != 0 {
		log.Printf("Identify(): response status code %d",
			echonestResponse.Response.Status.Code)
		return nil, errors.New(echonestResponse.Response.Status.Message)
	}

	if len(echonestResponse.Response.Songs) > 0 {
		log.Printf("Identify(): found %v", echonestResponse.Response.Songs)
		meta := echonestResponse.Response.Songs[0]
		return &api.Meta{
			Artist: meta.ArtistName,
			Title:  meta.Title,
		}, nil
	}

	return nil, nil
}
